/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: como_client.c,v 1.21 2005/01/21 17:20:43 lrizzo Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>     /* bzero */
#include <sys/types.h>  /* socket, connect */
#include <sys/socket.h> /* socket, connect */
#include <netinet/in.h> /* struct sockaddr_in */
#include <sys/time.h>   /* gettimeofday */
#include <unistd.h>     /* getopt */
#include <err.h>	/* err, errx */

#include "qutils.h"
#include "query.h"
#include "stdpkt.h"     /* TIME2TS */

static char helpstring[] = "\n\
A simple CoMo client that issues queries to a running server.\n\
The query is specified through the command line, and replies\n\
are printed out until the server closes the connection.\n\
";

static void
usage(const char *name, qreq_t *m)
{
	fprintf(stderr,
"Usage: %s [-f format] [-i addr]\n\
\t [-p port] [-c clname] [-f filter] [-s start] [-e end]\n\
\t-f: output format, as in printf (default %%d)\n\
\t-i: query server address (default localhost)\n\
\t-p: query server port (default %d)\n\
\t-c: classifier name (default \"%s\")\n\
\t-t: filter expression (default \"%s\")\n\
\t-s: start time (in sec, default 60s ago)\n\
\t-e: end time (in sec, default now)\n\
\t-h: display this help and exit\n\n",
	name,
	QSERV_PORT,
	m->module,
	m->filter);
}

/*
 * -- parse_cmdline
 */
static int
parse_cmdline(int argc, char *argv[], char *fmt,
	struct sockaddr_in *addr, qreq_t *m)
{
    struct timeval tv;
    int opt;

    /* default options */
    strcpy(fmt, "%d %lld %lld\n");
    bzero(addr, sizeof(addr));
    addr->sin_family = AF_INET;
    addr->sin_port = htons(QSERV_PORT);
    inet_aton("127.0.0.1", &addr->sin_addr);
    //inet_aton("10.212.4.131", &addr->sin_addr);

    /* fill default query request */
    m->type = Q_QUERY; 
    m->len = sizeof(qreq_t); 
    strcpy(m->module, "counter");
    strcpy(m->filter, "ALL");
    gettimeofday(&tv, NULL);
    m->start = tv.tv_sec;
    m->end = m->start + 10;  //tv.tv_sec + 60; 

    while ((opt = getopt(argc, argv, "f:i:p:c:t:s:e:h")) != -1) {
        switch (opt) {
	case 'f':   /* format */
	    snprintf(fmt, MAX_FMT, "%s", optarg);
	    break;
	case 'i':   /* query server address */
	    if (inet_aton(optarg, &addr->sin_addr) <= 0)
		perror("inet_aton");
	    break;
	case 'p':   /* query server port */
	    addr->sin_port = htons(atoi(optarg));
	    break;
	case 'c':   /* classifier name */
	    strncpy(m->module, optarg, MODULENAME_MAX); 
	    break;
	case 't':   /* filter expression */
	    strncpy(m->filter, optarg, FILTERSTR_MAX); 
	    break;
	case 's':   /* start time */
	    m->start = atoi(optarg); 
	    break;
	case 'e':   /* end time */
	    m->end = atoi(optarg);
	    break;
	case 'h':
	    fprintf(stderr, "%s", helpstring);
	    /* FALLTHROUGH */
	default:
	    usage(argv[0], m);
	    return -1;
        }
    }

    return 0;
}


size_t 
safe_read(int sd, char * buf, size_t len)
{
    int rd; 

    rd = read(sd, buf, len); 
    if (rd < 0)
        err(EXIT_FAILURE, "reading from socket"); 
    if (rd == 0) {
        /* connection closed gracefully (?!?!) */
        fprintf(stderr, "stream finished.");
        exit(0);
    }
    return (size_t) rd; 
}

/*
 * -- main
 */
int
main(int argc, char *argv[])
{
    char fmt[MAX_FMT];
    struct sockaddr_in addr;
    qreq_t m;
    int sd;

    /* check and get command line arguments */
    if (parse_cmdline(argc, argv, fmt, &addr, &m) < 0)
        exit(0);

    /* connect to query server */
    if ((sd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
        perror("socket");
        exit(0);
    }

    if (connect(sd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        perror("connect");
        close(sd);
        exit(0);
    }

    /* send query to query server */
    fprintf(stderr, 
	"request classifier [%s] filter [%s]\n"
	"\tstart %d size %d\n",
	m.module, m.filter, m.start, m.end - m.start);
    if (write(sd, &m, m.len) != m.len) {
        close(sd);
        perror("write_query");
        exit(0);
    }

    while (1) {
        char buf[4096];
	qctl_t * ctl = (qctl_t *) buf;
        qdata_t * data = (qdata_t *) buf;
	size_t rd, ofs, len;

        ofs = 0; 

	/* read the message type first */
	rd = safe_read(sd, buf, 1); 
	ofs += rd; 

	switch (buf[0]) { 
	case Q_ERR: 
	    len = sizeof(qctl_t); 
	
            while (ofs < len) {
                rd = safe_read(sd, &buf[ofs], len - ofs);
                ofs += rd;
	    } 

            fprintf(stderr, "ERROR %s\n", strerror(ctl->code)); 
            close(sd);
            exit(0);

	case Q_DATA: 
	    len = sizeof(qdata_t); 
	
            while (ofs < len) {
                rd = safe_read(sd, &buf[ofs], len - ofs);
                ofs += rd;

                /* if we have read enough of this message, 
	  	 * check the actual message length 
		 */
                if (ofs >= sizeof(qdata_t)) 
                    len = (size_t) ntohs(data->len);
	    }
		
            /* print record according to given format */
            qfprintf(stdout, fmt, data->rec); 
	    break;

	default: 
	    errx(EXIT_FAILURE, "unknown message (%d)\n", buf[0]); 
        }
    } 

    /* otherwise we assume that the message received was CLOSE */
    close(sd);

    return 0;
}
