/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: qutils.c,v 1.1 2004/09/02 17:30:18 pbarlet Exp $
 */

#include <unistd.h>
#include <string.h>     /* bzero */
#include <sys/types.h>  /* socket, bind, listen, accept */
#include <sys/socket.h> /* socket, bind, listen, accept */
#include <netinet/in.h> /* struct sockaddr_in */
#include <arpa/inet.h>  /* inet_ntoa */

#include "qutils.h"
#include "module.h"

/*
 * -- init_server
 *
 * init a query collector server and wait for a connection from
 * the query process
 *
 */
int
init_server(int port)
{
    int sfd, cfd, opt = 1;
    struct sockaddr_in saddr, caddr;
    socklen_t clen = sizeof(caddr);

    /* create socket to receive query results */
    if ((sfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        return -1;

    /* allow local address reuse in TIME_WAIT */
    if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
        close(sfd);
        return -1;
    }

    /* assign address to the socket */
    bzero(&saddr, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(port);
    saddr.sin_addr.s_addr = INADDR_ANY;
    if (bind(sfd, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
        close(sfd);
        return -1;
    }

    /* wait for a connection from query */
    if (listen(sfd, 1) < 0) {
        close(sfd);
        return -1;
    }

    /* accept connection from query */
    bzero(&caddr, sizeof(caddr));
    if ((cfd = accept(sfd, (struct sockaddr *)&caddr, &clen)) < 0) {
        close(sfd);
        return -1;
    }

    /* for the moment we only accept only one connection */
    close(sfd);

    /* return socket descriptor which connects query and collector */
    return cfd;
}


/*
 * -- qfprintf
 *
 * Read a record sent by query for sendto and on-demand queries, and
 * print it according to the format specified in "fmt". The format is
 * almost the same as printf, but includes some new features.
 * Unfortunately this ugly function is needed for so many things:
 *
 *  . integer values are converted from network into host byte order.
 *  . the format specifes the number of bytes consumed from "data",
 *    so record fields in "data" doesn't have to be word aligned. E.g.
 *    "hhu" read and print only 1 byte from data and advance the read
 *    pointer only one byte, while "%llu" would read 8 bytes and would
 *    advance the read pointer by 8.
 *  . print correclty escape characters like \n or \t. This can't be
 *    done calling directly to printf, since the translation of escape
 *    characters is done in compile time, but "fmt" is expected to be
 *    specified in the command line.
 *  . introduce a new format "%a" which convert an integer address to
 *    a string and print it.
 *  . "%-?" read the number of bytes specified by ? from "data" (where
 *    ? is any printf conversion specifier) but supress its output. It
 *    is similar to the "*" flag character of scanf. This is very useful
 *    if we don't want to print all the fields of a record sent by
 *    query. E.g. "%-d %d" will read two integers but only will print
 *    the second one.
 *
 * This function still needs a lot of work, but for the moment it seems
 * to understand most of printf formats. This function is not expected to
 * work with bad written format strings, but given a correct format string
 * it should print a correct output. So, if the format string is incorrect
 * print results are still unpredictable. For the moment only tested to
 * print integers...
 *
 */
void
qfprintf(FILE *fp, char *fmt, char *data)
{
    char *cp = data;
    char fma[MAX_FMT];
    struct in_addr addr;
    int i, len, printit;

    while (1) {
        while (*fmt != '\0' && *fmt != '\\' && *fmt != '%')
            fprintf(fp, "%c", *fmt++);    //XXX this can be done better

        switch(*fmt) {
            case '\\':
                switch(*++fmt) {
                    case 'n':
                        fprintf(fp, "\n");
                        break;
                    case 't':
                        fprintf(fp, "\t");
                        break;
                    case '\0':
                        return;
                    default:
                        fprintf(fp, "\\%c", *fmt);
                 }
                 break;
            case '%':
                printit = 1;
                fma[0] = '%';
                i = 0;
                len = 4;
lswtch:         fma[i++] = *fmt;
                switch (*++fmt) {
                    case 'l':
                        if (*(fmt-1) == 'l')    //XXX provisional
                            len <<= 1;
                        goto lswtch;
                    case 'h':
                        len >>= 1;
                        goto lswtch;
                    case 's':
                        while (*cp != '\0')
                            fprintf(fp, "%c", *cp++);
                        break;
                    case 'a':   /* %a print an ipv4 address */
                        addr.s_addr = *(uint32_t*)cp;
                        if (printit)
                            fprintf(fp, "%s", inet_ntoa(addr));
                        cp += 4;
                        break;
                    case '-':   /* supress the output for the current value */
                        printit = 0;
                        goto lswtch;
                    case 'e':   //XXX e,E,f,g,G not tested
                    case 'E':
                    case 'f':
                    case 'g':
                    case 'G':
                        len = 8;
                    case 'i':
                    case 'd':
                    case 'u':
                    case 'o':
                    case 'x':
                    case 'X':
                    case 'p':   //XXX p,D,O,U not tested (len should be 4)
                    case 'D':
                    case 'O':
                    case 'U':
                        fma[i++] = *fmt;
                        fma[i]   = '\0';
                        if (printit) {
                            switch (len) {
                                case 1:
                                    fprintf(fp, fma, *(uint8_t*)cp);
                                    break;
                                case 2:
                                    fprintf(fp, fma, ntohs(*(uint16_t*)cp));
                                    break;
                                case 4:
                                    fprintf(fp, fma, ntohl(*(uint32_t*)cp));
                                    break;
                                case 8:
                                    fprintf(fp, fma, NTOHLL(*(uint64_t*)cp));
                                    break;
                                default:
                                    return;
                            }
                        }
                        cp += len;
                        break;
                    case '%':
                        fprintf(fp, "%%");
                        break;
                    case '\0':
                        return;
                    default:
                        goto lswtch;

                }
                break;
            case '\0':
            default:
                return;
        }
        fmt++;
    }
}
