/*
 * Copyright (c) 2004, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: capture.c,v 1.60 2005/01/23 19:52:20 gianluca Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>	/* F_GETFL */
#include <unistd.h>     /* read, write etc. */
#include <string.h>     /* bzero */
#include <errno.h>      /* errno */

#include "como.h"
#include "sniffers.h"

/* poll time (in usec) */
#define POLL_WAIT   1000000

/* global state */
extern struct _como map;

/* sniffer list and callbacks */
extern struct _sniffer *__sniffers[];

/*
 * Support for tailq handling.
 * Used for the expired tables.
 */
typedef struct {
    void * __head;
    void * __tail;
} tailq_t;


#define TQ_HEAD(queue)  ((queue)->__head)

#define TQ_APPEND(queue, entry, link_field)             \
    do {                                                \
        tailq_t *q = (queue);                           \
        typeof(entry) e = entry;                        \
        if (q->__head)                                  \
            ((typeof(entry))q->__tail)->link_field = e; \
        else                                            \
            q->__head = e;                              \
        q->__tail = e;                                  \
        e->link_field = NULL;                           \
    } while (0);



/*
 * -- create_filter
 *
 * creates a filter on the fly. it uses the template provided
 * during the configuration to generate the source code for the
 * filter directly from the expression contained in each node
 * of all the modules. it returns the file name.
 *
 */
static char *
create_filter(module_t * mdl, int count, char * template)
{
    char *buf;
    FILE * fp;
    int ret;

    asprintf(&buf, "%s/my_filter", map.workdir);
    fp = fopen(buf, "w");
    if (fp == NULL)
        panic("cannot open temp file %s\n", buf);
    free(buf);

    fprintf(fp, "#define MODULE_COUNT %d\n", count);
    fprintf(fp, "#define USERFILTER \\\n");

    for (; mdl; mdl = mdl->next) 
         fprintf(fp, "\touts[%d][i] = (%s);\\\n", mdl->index, mdl->filter);

    fprintf(fp,"\n");

    fclose(fp);

    asprintf(&buf,
        "(cd %s; cc -O -I- -I . -nostdlib -shared -Wl,-x,-S \\\n"
        "\t\t-o my_filter.so -include my_filter %s )",
	map.workdir, template);
    logmsg(LOGCAPTURE, "compiling filter \n\t%s\n", buf);
    ret = system(buf);
    free(buf);
    if (ret != 0) {
	logmsg(LOGWARN, "compile failed with error %d\n", ret);
	return NULL;
    }
    /* XXX TODO: check for compile failure */
    asprintf(&buf, "%s/my_filter.so", map.workdir);
    return buf;
}


/* 
 * -- create_table 
 * 
 * allocates and initializes a hash table
 */
ctable_t * 
create_table(module_t * mdl, timestamp_t ts) 
{
    ctable_t * ct; 
    size_t len;

    len = sizeof(ctable_t) + mdl->ca_hashsize * sizeof(void *);
    ct = new_mem(NULL, len, "new_flow_table");
    if (ct == NULL) 
	panic("new_mem failed"); 

    ct->size = mdl->ca_hashsize;
    ct->first_full = ct->size;      /* all records are empty */
    ct->module = mdl;
    ct->ts = ts; 

    // XXX what is this for? 
    ct->mem = NULL;                 /* use the system's map */

    return ct; 
}


/*
 * -- flush_table
 *
 * Called by capture_pkt() process when a timeslot is complete.
 * it flushes the flow table (if it exists and it is non-empty)
 * and then it allocates a new one.
 *
 */
static void
flush_table(module_t * cl, tailq_t * expired)
{
    ctable_t *ct;

    logmsg(V_LOGCAPTURE, "flush_table start\n");

    /* check if the table is there and if it is non-empty */
    ct = cl->ca_hashtable;
    if (ct && ct->records) {
        if (ct->records > ct->size)
            logmsg(LOGWARN,
		"flush_table table '%s' overfull (%d vs %d) -- %d live\n",
		cl->name, ct->records, ct->size, ct->live_buckets);

        logmsg(V_LOGCAPTURE,
            "flush_tables %p(%s) buckets %d records %d live %d\n", ct,
            cl->name, ct->size, ct->records, ct->live_buckets);

        /* add to linked list and remove from here. */
        TQ_APPEND(expired, ct, next_expired);
        cl->ca_hashtable = NULL;
    }

    logmsg(V_LOGCAPTURE, "classifier %s flush_table done.\n", cl->name);
}


/*
 * -- capture_pkt
 *
 * This function is called for every batch of packets that need to be
 * processed by a classifier.
 * For each packet in the batch it runs the check()/hash()/match()/update()
 * methods of the classifier cl_index. The function also checks if the
 * current flow table needs to be flushed.
 *
 */
static void
capture_pkt(int idx, pkt_t * pkt, int no_pkts, int * which, tailq_t * expired)
{
    module_t * cl;
    ctable_t * ct; 
    int i;

    cl = map.mdl_array[idx];
    if (cl->ca_hashtable == NULL) 
	cl->ca_hashtable = create_table(cl, pkt->ts); 

    ct = cl->ca_hashtable;

    for (i = 0; i < no_pkts; i++, pkt++) {
        rec_t *prev, *cand;
        uint32_t hash;
        uint bucket;

        /* flush the current flow table, if needed */
        if (pkt->ts > ct->ts + cl->ca_ivl) {  
            flush_table(cl, expired);
            ct = cl->ca_hashtable = create_table(cl, pkt->ts); 
	}

        if (which[i] == 0)
            continue;   /* no interest in this packet */

        /* unset the filter for this packet */
        which[i] = 0;

        /*
         * check if there are any errors in the packet that
         * make it unacceptable for the classifier.
         * (if check() is not provided, we take the packet anyway)
         */
        if (cl->callbacks.check != NULL && !cl->callbacks.check(pkt))
            continue;

        /*
         * find the entry where the information related to
         * this packet reside
         * (if hash() is not provided, it defaults to 0)
         */
        hash = cl->callbacks.hash != NULL ? cl->callbacks.hash(pkt) : 0;
        bucket = hash % ct->size;

        /*
         * keep track of the first entry in the table that is used.
         * this is useful for the EXPORT process that will have to
         * scan the entire hash table later.
         */
        if (bucket < ct->first_full)
            ct->first_full = bucket;

        prev = NULL;
        cand = ct->bucket[bucket];
        while (cand) {
            /* if match() is not provided, any record matches */
            if (cl->callbacks.match == NULL || cl->callbacks.match(pkt, cand))
                break;
            prev = cand;
            cand = cand->next;
        }

        if (cand != NULL) {
           /*
            * found!
            * two things to do first:
            *   i) move this record to the front of the bucket to
            *      speed up future (and likely) accesses.
            *  ii) check if this record was flagged as full and
            *      in that case create a new one;
            */

            /* move to the front, if needed */
            if (ct->bucket[bucket] != cand) {
                prev->next = cand->next;
                cand->next = ct->bucket[bucket];
                ct->bucket[bucket] = cand;
            }

            /* check if this record was flagged as full */
            if (cand->full) {
                rec_t * x;

                /* allocate a new flow record */
                x = new_mem(ct->mem, cl->callbacks.ca_recordsize, "new");
                x->hash = cand->hash;
                x->next = cand->next;

                /* link the current full one to the list of full records */
                x->prev = cand;
                cand->next = x;

                /* we moved cand to the front, now is x */
                ct->bucket[bucket] = x;

                /* done. new empty record ready. */
                /* NOTE: it is correct not to increment ct->records here */

                /* now update the new record */
                x->full = cl->callbacks.update(pkt, x, 1);
                continue;
            }

            /* not full and not new, update the record */
            cand->full = cl->callbacks.update(pkt, cand, 0);
            continue;
        }

        /*
         * not found!
         * create a new record, update flow table stats
         * and link it to the bucket.
         */
        cand = new_mem(ct->mem, cl->callbacks.ca_recordsize, "new");
        cand->hash = hash;
        cand->next = ct->bucket[bucket];

        ct->records++;
        ct->bucket[bucket] = cand;
	if (cand->next == NULL) 
	    ct->live_buckets++;

        /* now update the new record */
        cand->full = cl->callbacks.update(pkt, cand, 1);
    }
}

struct _sniffers_list {
    struct _sniffers_list *next;
    int	fd;
    sniffer_t	cb;	/* callbacks */
};

/*
 * -- capture_mainloop
 *
 * This is the CAPTURE mainloop. It first prepares the filter
 * (that could be compiled on the fly) and then opens the sniffer
 * device. Then the real mainloop starts and it sits on a select
 * waiting for messages from EXPORT, the SUPERVISOR or for packets
 * from the sniffer.
 *
 */
void
capture_mainloop(int export_fd)
{
    filter_fn * filter;     /* filter function */
    memlist_t * flush_map;  /* freed blocks in the flow tables */
    tailq_t expired;        /* expired flow tables */
    struct _sniffers_list *sn, *devs = NULL;
    pkt_t pkts[MAX_PKTS];   /* batch of packets from sniffer */
    int sniffers_left = 0;  /* how many sniffers are left ? */
    int sent = 0;           /* message sent to export */

    /* we only get a socketpair to the export process */

    /*
     * load the filter in. if no object has been provided, then
     * the filter must be compiled on the fly. we will write the
     * filter on a temporary file, call the compiler and finally
     * link in the new shared object.
     */

    if (!map.filter)
        map.filter = create_filter(map.modules, map.modules_count, map.template);

    filter = load_object(map.filter, "filter");

    /* get the sniffer callbacks and start it. */
    /* XXX todo - build a list of sniffers we use */
    for (sniffers_left = 0;;) {
	sn = safe_calloc(1, sizeof(*sn));
	sn->cb = *__sniffers[map.sniffer];
	sn->fd = sn->cb.sniffer_start(map.device, map.check_direction);
	if (sn->fd < 0)
	    panic("error opening capture device (%s)\n", map.device);
	sn->next = devs;
	devs = sn;
	sniffers_left++;

	break;
    }
    /*
     * allocate memory used to pass flow tables
     * between CAPTURE and EXPORT
     */
    flush_map = new_memlist(16);

    /* init expired flow tables queue */
    TQ_HEAD(&expired) = NULL;

    /*
     * This is the actual main loop where we monitor the various
     * sniffers and the sockets to communicate with other processes.
     * If a sniffer's data stream is complete or fails, we close it.
     * The loop terminates when all sniffers are closed and there is
     * no pending communication with export.
     */
    for (;;) {
	fd_set r;
	int i, maxfd = 0;
	struct timeval t = {POLL_WAIT / 1000000, POLL_WAIT % 1000000};
	struct timeval *pto = NULL;

        errno = 0;

	if (sniffers_left == 0 && !sent && TQ_HEAD(&expired) == NULL)
	    break;	/* nothing more to do */

	FD_ZERO(&r);
	
	if (sent) {	/* we talked to export, so listen for the reply */
	    FD_SET(export_fd, &r);
	    maxfd = export_fd;
	}
	for (sn = devs; sn; sn = sn->next) {
	    if (sn->fd < 0 || (sn->cb.flags & SNIFF_POLL)) {
		pto = &t; 
		continue;	/* do not select on this one */
	    }
	    FD_SET(sn->fd, &r);
	    if (sn->fd > maxfd)
		maxfd = sn->fd;
	}

	/* wait for messages, sniffers or up to the polling interval */
	i = select(maxfd+1, &r, NULL, NULL, pto);
	if (i < 0)
	    panic("select: %s\n", strerror(errno));

        if (FD_ISSET(export_fd, &r)) {	/* export replies to our message */
	    int ret;
	    msg_t reply;

	    errno = 0;	/* reset */
	    ret = read(export_fd, &reply, sizeof(reply));
	    if (ret != 0 && errno != EAGAIN) {
		if (ret != sizeof(reply))
		    panic("**** error reading export_fd got %d\n", ret);

		if (reply.m != flush_map)
		    panic("**** bad flush_map from export_fd\n");

		/* ok, export freed memory into the map for us */
		ret = mem_merge_maps(NULL, flush_map);
		sent = 0;   /* mark no pending jobs */
	    }
        }

        /* need to write and have no pending jobs from export */
        if (!sent && (TQ_HEAD(&expired)) != NULL) {
	    msg_t x;	/* message to EXPORT */
            x.m  = flush_map;
            x.ft = TQ_HEAD(&expired);
            if (write(export_fd, &x, sizeof(x)) != sizeof(x))
                panic("**** error writting export_fd\n");

            TQ_HEAD(&expired) = NULL;   /* we are done with this. */
            sent = 1;                   /* wait response from export */
        }

	if (sniffers_left == 0)
	    continue;

        /* check sniffers for packet reception */
	for (sn = devs; sn; sn = sn->next) {
            int * which;
            int idx;
	    int count = 0;

	    if (sn->fd < 0)
		continue;	/* invalid device */

	    if ( !(sn->cb.flags & SNIFF_POLL) && !FD_ISSET(sn->fd, &r))
		continue;	/* nothing to read here. */

	    count = sn->cb.sniffer_next(sn->fd, pkts);
	    if (count == 0)
		continue;

            if (count < 0) {
                sn->cb.sniffer_stop(sn->fd);
		sn->fd = -1;
		sniffers_left--;
                continue;
            }

            /*
             * Select which classifiers need to see which packets
	     * The filter() function (see comments in file base/template)
	     * returns a bidimensional array of integer which[cls][pkt]
	     * where the first index indicates the classifier, the second
	     * indicates the packet in the batch.
	     * The element of the array is set if the packet is of interest
	     * for the given classifier, and it is 0 otherwise.
             */
            which = filter(pkts, count, map.modules_count);

            /*
             * Now browse through the classifiers and
             * perform the capture actions needed.
             *
             * XXX we do it this way just because anyway we
             *     have got one CPU... will have to change in
             *     a really distributed system. one day.
             *
             */

            for (idx = 0; idx < map.modules_count; idx++) {
                capture_pkt(idx, pkts, count, which, &expired);
                which += count; /* next classifier, new list of packets */
            }
        }
    }

    logmsg(LOGWARN, "capture: sniffer has been closed.");
    return;
}
/* end of file */
