/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: config.c,v 1.38 2005/01/25 21:28:02 gianluca Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h> 	/* bcmp */
#include <ctype.h>  	/* isalnum */
#include <err.h>    	/* warnx, errx */
#include <dirent.h>	/* opendir */
#include <sys/types.h>	/* mkdir, opendir */
#include <sys/stat.h>	/* mkdir */

#include <errno.h>

#include "como.h"
#include "sniffers.h"

/* global data structure that contains all
 * configuration information
 */
extern struct _como map;


/* get the sniffers list */
extern sniffer_t *__sniffers[];

/*
 * tokens used in the dictionary
 */
enum tokens {
    TOK_NULL=0,

    TOK_BASEDIR,
    TOK_BLOCK_SIZE,
    TOK_SOURCE,
    TOK_DESCRIPTION,
    TOK_END,
    TOK_FILESIZE,
    TOK_FILTER,
    TOK_HASHSIZE,
    TOK_LIBRARYDIR,
    TOK_LOGFLAGS,
    TOK_MEMSIZE,
    TOK_MODULE,
    TOK_OUTPUT,
    TOK_SNIFFER,
    TOK_VERBOSE
};


/*
 * in the configuration file there are different scopes, or contexts.
 * the scope of a configuration variable could be global (i.e., default
 * values that affect all objects), or related to a specific module,
 * classifier or query.
 */
#define CTX_ANY         0xff
#define CTX_GLOBAL      0x01
#define CTX_MODULE      0x02

/*
 * this structure is a dictionary of symbols that can be
 * found in the configuration file.
 *
 * _keyword is a structure that stores a string <-> token pairs, used in
 * various places in the parser. Entries are stored in arrays,
 * with an entry with s=NULL as terminator.
 * The search routines are match_token() and match_value().
 * Often, an element with x=0 contains an error string.
 *
 */

struct _keyword {
    char const *str;    /* keyword */
    int action;         /* related token */
    int nargs;          /* arguments */
    int scope;          /* specific scope */
};

typedef struct _keyword keyword_t;

keyword_t keywords[] = {
    { "basedir",     TOK_BASEDIR,     2, CTX_GLOBAL },
    { "librarydir",  TOK_LIBRARYDIR,  2, CTX_GLOBAL },
    { "logflags",    TOK_LOGFLAGS,    2, CTX_GLOBAL },
    { "module",      TOK_MODULE,      2, CTX_GLOBAL },
    { "sniffer",     TOK_SNIFFER,     3, CTX_GLOBAL },
    { "verbose",     TOK_VERBOSE,     1, CTX_GLOBAL },
    { "memsize",     TOK_MEMSIZE,     2, CTX_GLOBAL | CTX_MODULE},
    { "output",      TOK_OUTPUT,      2, CTX_MODULE},
    { "blocksize",   TOK_BLOCK_SIZE,  2, CTX_MODULE},
    { "filesize",    TOK_FILESIZE,    2, CTX_MODULE},
    { "hashsize",    TOK_HASHSIZE,    2, CTX_MODULE},
    { "source",      TOK_SOURCE,      2, CTX_MODULE},
    { "filter",      TOK_FILTER,      2, CTX_GLOBAL|CTX_MODULE},
    { "description", TOK_DESCRIPTION, 2, CTX_MODULE},
    { "end",         TOK_END,         1, CTX_ANY },
    { NULL,          0,               0, 0 }    /* terminator */
};


/*
 * -- match_token
 *
 * takes the keywords table and a string and returns
 * the value associated with the string in the table.
 *
 */
static keyword_t *
match_token(char * string, keyword_t *table)
{
    keyword_t *pt;
    uint i = strlen(string);

    for (pt = table ; i && pt->str != NULL ; pt++)
        if (strlen(pt->str) == i && !bcmp(string, pt->str, i))
            return pt;
    return NULL;
}

/*
 * -- do_config
 *
 * Do the actual configuration.
 * Expect each useful line to start with a keyword.
 *
 */
static void
do_config(int argc, char *argv[])
{
    static int scope = CTX_GLOBAL;      /* scope of current keyword */
    static module_t * mdl = NULL;       /* module currently open */
    keyword_t *t;
    callbacks_t *cb;
    char *clname;
    int i;

    /*
     * run some checks on the token (i.e., that it exists
     * and that we have all the arguments it needs).
     */
    t = match_token(argv[0], keywords);
    if (t == NULL)
        panic("unknown token \"%s\"\n", argv[0]);

    if (argc < t->nargs)
        panic("\"%s\" requires at least %d arguments\n", argv[0], t->nargs);
    if (!(t->scope & scope))
        panic("\"%s\" out of scope\n", argv[0]);


    switch (t->action) {
    case TOK_BASEDIR:
  	if (map.basedir)		/* get rid of previous configs */
	    free(map.basedir);
	map.basedir = strdup(argv[1]); 
	if (opendir(map.basedir) == NULL) {
	    int ret; 

	    ret = mkdir(map.basedir, (mode_t) (S_IRWXU | S_IRWXG | S_IRWXO));
	    if (ret < 0) 
		panic("cannot create basedir %s: %s\n", 
		    map.basedir, strerror(errno)); 
	}
	break;

    case TOK_BLOCK_SIZE:
	mdl->bsize = atoi(argv[1]);
	break;

    case TOK_DESCRIPTION:
	mdl->description = strdup(argv[1]);
	break;

    case TOK_END:
        /*
         * "end" of a module configuration.  run some checks depending 
         * on context to make sure that all mandatory fields are there
         * and set default values
         */
        if (scope == CTX_MODULE) {
            if (mdl->filter == NULL && map.filter == NULL)
                panic("config: module %s needs a filter\n", mdl->name);
            if (mdl->output == NULL)
                panic("config: module %s needs an output\n", mdl->name);
            if (mdl->ex_hashsize == 0)
                panic("config: module %s needs a hash size\n", mdl->name);

            /* perform some checks on the callbacks */
	    cb = &mdl->callbacks;

            /* update(), store() and load() should definitely be there */
            if (cb->update == NULL)
                panic("source: module %s misses update()\n", mdl->name);
            if (cb->store == NULL)
                panic("source: module %s misses store()\n", mdl->name);
            if (cb->load == NULL)
                panic("source: module %s misses load()\n", mdl->name);

            /* 
	     * either both or none of action() and export() are required 
	     */
            if (cb->export == NULL) { /* no export() ... */
		/* we don't want action() either */
		if (cb->action != NULL)
		    panic("source: module %s has action() without export()\n",
			 mdl->name);

		/* ematch() and compare() are useless */
                if (cb->ematch != NULL)
		    panic("source: module %s has ematch() w/out export()\n", 
			mdl->name);
                if (cb->compare != NULL)
		    panic("source: module %s compare() w/out export()\n", 
			mdl->name);
            } else {
		if (cb->action == NULL)
		    panic("source: module %s has export() w/out action()\n", 
			mdl->name);
	    }

	    /* initiate classifier */
	    if (cb->init != NULL)
		if (cb->init(mdl->mem, mdl->msize) != 0)
		    panic("could not initialize %s\n", mdl->name); 

	    logmsg(LOGUI, "... loaded module \"%s\"\n", mdl->name); 
	    if (mdl->description != NULL) 
		logmsg(LOGUI,"     Description: %s\n", mdl->description); 
	    logmsg(LOGUI,"     Filter: %s\n", mdl->filter); 
	    logmsg(LOGUI,"     Output: %s\n", mdl->output); 

            /* change scope */
            scope = CTX_GLOBAL;
        } else if (scope == CTX_GLOBAL) {
            /* nothing to do */
        }
        break;

    case TOK_FILESIZE:
	mdl->filesize = atoi(argv[1]);
	break;

    case TOK_FILTER:
	if (scope == CTX_MODULE) {
	    mdl->filter = strdup(argv[1]);
	} else if (scope == CTX_GLOBAL) {
	    if (map.filter)
		free(map.filter);
	    map.filter = strdup(argv[1]);
	}
	break;

    case TOK_HASHSIZE:
        mdl->ex_hashsize = atoi(argv[1]);
        mdl->ca_hashsize = mdl->ex_hashsize; 
	/* XXX we would like to make the capture hashsize adaptive. 
	 *     in general it is going to be much smaller than the 
	 *     one used in export. 
	 */
    
        /* 
         * If this is the first hashsize keyword 
         * initialize the export hash table and the array of records.
         */
        if (mdl->ex_hashtable == NULL) {
            size_t len; 
    
            /* allocate hash table */
            len = sizeof(etable_t) + mdl->ex_hashsize * sizeof(void *);
            mdl->ex_hashtable = safe_calloc(1, len); 
            mdl->ex_hashtable->size = mdl->ex_hashsize; 
    
            /* allocate record array */
            len = sizeof(earray_t) + mdl->ex_hashsize * sizeof(void *);
            mdl->ex_array = safe_calloc(1, len); 
            mdl->ex_array->size = mdl->ex_hashsize; 
        }
        break;

    case TOK_SOURCE:
        /* load the library */
        if (!map.libdir || argv[1][0] == '/')
            clname = strdup(argv[1]);
        else
            asprintf(&clname, "%s/%s", map.libdir, argv[1]);

        cb = load_object(clname, "callbacks");
        if (cb == NULL)
            panic("could not load library %s\n", clname); 
        free(clname);

        /* store the callbacks */
        mdl->callbacks = *cb;
        break;

    case TOK_LIBRARYDIR:
        if (map.libdir)
            free(map.libdir);
        map.libdir = strdup(argv[1]);
        break;

    case TOK_LOGFLAGS:  /* XXX simplify this interface... */
        map.logflags = strtoul(argv[1], NULL, 0);
        break;

    case TOK_MEMSIZE:
        /* this keyword can be used in two contexts */
        if (scope == CTX_GLOBAL) {
            map.mem_size = atoi(argv[1]);
            if (map.mem_size <= 0 || map.mem_size > 256)
                panic("invalid memory size %d, range is 1..256\n", 
		    map.mem_size);
        } else if (scope == CTX_MODULE) {
            /* private module memory */
            mdl->mem = safe_calloc(1, (size_t) atoi(argv[1])); 
            mdl->msize = atoi(argv[1]);
        }
        break;

    case TOK_MODULE:
        /* new module. need to allocate a new module element */
        mdl = safe_calloc(1, sizeof(module_t));
        mdl->name = strdup(argv[1]);
        mdl->next = map.modules;
        map.modules = mdl;
        map.modules_count++;

	/* set some default values */
	mdl->ca_ivl = TIME2TS(1,0);
	mdl->bsize = DEFAULT_BLOCKSIZE;
	mdl->filesize = DEFAULT_FILESIZE; 

        /* change scope */
        scope = CTX_MODULE;
        break;

    case TOK_OUTPUT:
	if (!map.basedir || argv[1][0] == '/')
	    mdl->output = strdup(argv[1]);
	else
	    asprintf(&mdl->output, "%s/%s", map.basedir, argv[1]);
	break;

    case TOK_SNIFFER:
        for (i = 0; __sniffers[i]; i++) {
	    const char *name = __sniffers[i]->name;
            if (bcmp(argv[1], name, strlen(name)) == 0) {
                map.sniffer = i;
                map.device = strdup(argv[2]);
                break;
            }
        }
        if (__sniffers[i] == NULL)
            panic("unknown source \"%s\"\n", argv[1]);

	map.check_direction =
        	(argc < 4) ? 1 : bcmp(argv[3], "nodir", strlen(argv[3]));
        break;

    case TOK_VERBOSE:
	map.logflags |= (map.logflags << 16); 
	break;

    default:
	logmsg(LOGWARN, "unknown keyword %s\n", argv[0]);
	break;
    }
}


/*
 * -- parse_cfgline
 *
 * parses a line in the configuration file. the way it operates
 * is simple: split the line into NULL-separated keywords and then
 * call do_config() to actually update the global "map".
 * note that this function handles multiple lines as well.
 * it keeps quite a lot of state between calls in order to process
 * the multiple lines as a single one.
 *
 */
void
parse_cfgline(const char *line)
{
    static enum {
	IN_BLANK,
	IN_WORD,
	IN_QUOTES,
	END_QUOTES,
	DONE
    } state = DONE;
    static int linenum;
    static uint dst;	/* output buffer, size and pointers */
    static uint len;
    static char *buf;

    int argc = 0;
    uint i, srclen;

#define EOL		"\n\r"		/* end of line */
#define WHITESP		" \t\f\v\n\r"	/* whitespace */
#define COMMENT		"#"		/* comment markers */
#define QUOTE		"\""		/* quote markers */

    srclen = strlen(line);
    linenum++;
    /*
     * check for any leftover from the previous round.
     */
    if (state == DONE) {
	dst = 0;
	state = IN_BLANK;
    }
    logmsg(V_LOGCONFIG, "parse [%3d] [%s]\n", linenum, line);
    for (i=0; i < srclen && state != DONE ; i++) {
	char c = line[i];
	int copy = 0;		/* must copy this character */
	int end_token = 0;	/* and also this is the end of a token */

	if (c == '\0') {
	    logmsg(LOGWARN, "unexpected end of line\n");
	    break;
	}
	switch (state) {
	case DONE:	/* we'll never get here, anyways */
	    break;

	case IN_BLANK:
	    /* we are seeing blanks. Just skip until next keyword */
	    if (index(WHITESP, c))	/* keep skipping spaces */
		break;
	    if (index(COMMENT, c)) {	/* end of line */
		state = DONE;
		break;
	    }
	    if (index(QUOTE, c)) {	/* start quote */
		state = IN_QUOTES;
		break;
	    }
	    if (isalnum(c)) {
		copy = 1;
		state = IN_WORD;
		break;
	    }
	    if (c == '\\' && i == srclen-1)	/* line cont */
		return;
	    logmsg(LOGCONFIG, "invalid char in line %d [%s]\n",
		linenum, line+i);
	    dst = 0;
	    return; /* error */

	case IN_WORD:
	    /* in a keyword. Keep reading until done. */
	    if (index(WHITESP, c)) {
		end_token = 1;
		state = IN_BLANK;
		break;
	    }
	    if (index(COMMENT, c)) {	/* end of line */
		end_token = 1;
		state = DONE;
		break;
	    }
	    if (index(QUOTE, c)) {	/* start quote */
		logmsg(LOGCONFIG, "invalid char in line %d [%s]\n",
		    linenum, line+i);
		dst = 0;
		return; /* error */
	    }
	    /* assume all the rest is valid (could be more restrictive). */
	    copy = 1;
	    break;

	case IN_QUOTES:
	    /* wait for end-quote */
	    if (index(QUOTE, c)) {	/* end quote */
		state = END_QUOTES;
		break;
	    }
	    /* anything is fine here */
	    copy = 1;
	    break;

	case END_QUOTES:
	    end_token = 1;
	    state = IN_BLANK;
	    break;
	}
	if (copy)
	    buf = add_char(buf, c, &dst, &len);
	if (end_token) {
	    buf[dst++] = '\0';
	    argc++;
	}
    }
    if (dst > 0 && buf[dst-1] == '\\') {
	buf = add_char(buf, '\n', &dst, &len);
	return;
    }
    switch(state) {
    case IN_QUOTES:
	logmsg(LOGCONFIG, "missing endquote in line %d [%s]\n",
	    linenum, line);
    	dst = 0;
	return; /* error */

    case END_QUOTES:
    case IN_WORD:
	buf = add_char(buf, '\0', &dst, &len);
	argc++;
	/* FALLTHROUGH */
    case DONE:
    case IN_BLANK:
	state = DONE;
	break;
    }
    /* now we have NUL-separated keywords. */
    if (argc == 0) {
	logmsg(V_LOGCONFIG, "empty line\n");
    } else {
	char **argv = malloc(argc * sizeof(char *));
	if (argv == NULL)
	    panic("cannot allocate buffer for line %d [%s]\n",
		    linenum, line);
	logmsg(V_LOGCONFIG, "");	/* start print */
	for (i=0, argc = 0; i<dst;i++) {
	    argv[argc++] = &buf[i];
	    logmsg(0, "<%s> ", buf+i);
	    while (i<dst && buf[i] != '\0')
		i++;
	}
	logmsg(0, "\n");
	do_config(argc, argv);
	free(argv);
    }
}

const char * usage =
    "usage: %s [-c config_file] -f pcap_file|-d bpf_device [-v logflags] [-m mem_size]\n";


/*
 * -- parse_config_file
 *
 * it opens the file received as input and parses one
 * line at a time calling parse_cfgline().
 *
 */
void
parse_cfgfile(char * name)
{
    char buf[2048];
    FILE *f;

    f = fopen(name, "r");
    if (f == NULL) {
	logmsg(LOGWARN, "cannot open config file (%s): %s\n",
	    name, strerror(errno));
	return;
    }

    while (fgets(buf, sizeof(buf), f)) {
        int l;

	/* get rid of last carriage return to handle multiple
	 * lines (i.e., lines that end with '\').
	 */
	l = strlen(buf);
	if (l > 0 && buf[l-1] == '\n')
	    buf[l-1] = '\0';

	parse_cfgline(buf);
    }

    fclose(f);
}


/*
 * use the malloc debugger (in FreeBSD)
 */
#if __FreeBSD__ == 4
extern char *malloc_options;
#elseif __FreeBSD__ == 5
extern char *_malloc_options;
#endif

/*
 * -- parse_cmdline
 *
 * parses the command line options and fills in the "map" (global)
 * data structure. some of the configuration parameters have already
 * been set with the default config file. command line options
 * overrides previous settings.
 *
 */
void
parse_cmdline(int argc, char *argv[])
{
    int c;

    opterr = 0;
    while ((c = getopt(argc, argv, ":c:d:s:m:vx:")) != -1) {

        switch(c) {
	case 'x':
	    /* pass debug options into a string */
	    if (strstr(optarg,"malloc=") == optarg) {
#if __FreeBSD__ == 4
		malloc_options = strdup(optarg+7);
#elseif __FreeBSD__ == 5
		_malloc_options = strdup(optarg+7);
#endif
		break;
	    }
	    if (map.debug) {
		char *old = map.debug;
		asprintf(&map.debug, "%s %s", old, optarg);
		free(old);
	    } else
		map.debug = strdup(optarg);
	    break;

        case 'c':   /* change config file name */
            parse_cfgfile(optarg);
            break;

        case 'd':   /* device/file name */
            map.device = strdup(optarg);
            break;

        case 's':   /* sniffer */
            map.sniffer = atoi(optarg);
            break;

        case 'm':   /* capture/export memory usage */
            map.mem_size = atoi(optarg);
            if (map.mem_size <= 0 || map.mem_size > 512)
                panic("invalid memory size %d, range is 1-256\n", map.mem_size);
            break;

        case 'v':   /* verbose */
            map.logflags |= (map.logflags << 16);
            break;

        case '?':   /* unknown */
            warnx("unrecognized option (%s)\n", argv[optind]);
            errx(EXIT_FAILURE, usage, argv[0]);
            break;

        case ':':   /* missing argument */
            warnx("missing argument for option (%s)\n", argv[optind]);
            break;

        default:    /* should never get here... */
            errx(EXIT_FAILURE, usage, argv[0]);
        }
    }
}
