/*-
 * Copyright (c) 2004, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Memory management functions.
 *
 * $Id: memory.c,v 1.4 2005/01/23 19:52:20 gianluca Exp $
 */

#include <stdlib.h>
#include <unistd.h>     /* usleep */
#include <sys/types.h>
#include <sys/mman.h>   /* mmap   */
#include <string.h>     /* bzero  */

#include "como.h"

#define MIN_SIZE	32	/* do not split less than this */

/*
 *

--- ARCHITECTURE ---

The memory allocator manages chunks of memory of variable size,
each described by a 'mem_block_t'. 'size' contains
the size of the user-portion (data[]) of the block, while
'next' is used to build free lists of blocks, typically of the same size.

Each list is reachable through a 'memlist_t' structure, which
normally contains a pointer to the head of the list, the size
of each entry, and the number of entries in the list.

A memory map is made of an array of such memlist_t's.
Entry #0 is special and used as follows:
  size	is the number of entries in the array;
  count	is the index of the first free entry;
  next	is a pointer to a list of blocks that could not
	fit in the arrays otherwise.
	THE BLOCKS IN THIS LIST MAY HAVE DIFFERENT SIZES.
Entries #1 .. #count-1 contain non-empty lists.
Entries #count .. #size-1 are empty.

The memory allocator functions uses a map as the first parameter.
NULL indicates a special map, private to the process,
which is pointed to by the static variable mem_map.
This map can be reallocated in case it
needs to be extended, and IS NOT in the shared memory segment.

Other maps cannot be reallocated, so when they run out of slots it
is necessary to store elements in entry #0, unsorted.

All allocations are done from maps, and presumably in a shared
memory segment.

Externally visible functions are:

  void * new_mem(memlist_t *map, int size, char *msg)
	returns a block of the desired size from the map;

  void free_mem(memlist_t *map, void *pointer);
	frees the block returning it to the map

  int merge_maps(memlist_t *dst, memlist_t *src);
	merges 'src' into 'dst', invalidating fields in 'src'.
	returns the amount of memory recovered.

  memlist_t *new_memlist(int entries)
	returns a new map (to be freed with free_mem)
 *
 */

typedef struct _mem_block mem_block_t;

struct _mem_block {
    struct _mem_block *next;
    void *_magic;
#define	MY_MAGIC	(void *)0x91919191
#define	MY_MAGIC_IN_USE	(void *)0x50b50b
    uint size;			/* actual size of data[] */
    char data[0];
};

/*
 * Entries of a memlist_t.
 */
struct _memlist {
    mem_block_t *next;
    uint size;			/* user data size */
    uint count;			/* elements in this list */
#define overflow_list(x)	(x)[0].next
#define	map_size(x)		(x)[0].size
#define	first_free(x)		(x)[0].count
};

/*
 * The min and max allowed shared addresses.
 * Any pointers should be in this range.
 */
static char *shared_low, *shared_high;

#define CHECK_PTR1(x, s) CHECK_PTR((mem_block_t *)(x) - 1, s)
#define	CHECK_PTR(x, s)						\
    do {							\
	char *c = (char *)x;					\
    if (c < shared_low || c >= shared_high)			\
	panic("%s corrupt ptr %p (should be %p.. %p\n",		\
		s, x, shared_low, shared_high);			\
    if ((x)->_magic != MY_MAGIC && (x)->_magic != MY_MAGIC_IN_USE)	\
	panic("%s bad magic at %p (%p should be %p)\n",		\
		s, x, (x)->_magic, MY_MAGIC);			\
    } while (0);

static memlist_t *mem_map;		/* the main map */
int	memory_in_use, peak;		/* diagnostic vars */

#if 1
#define	record_alloc(size, count, msg)	do {} while (0);
#else
/*
 * debugging only:
 *	size > 0, count = 0		record allocation.
 *	size > 0, count > 0		record deallocation of 'count' entries
 *	size = 0, count = 0		print allocation if map.verbose
 *	size = 0, count > 0		print allocation, forced.
 */
extern struct _como map;
static memlist_t *alloc_map;
static void
record_alloc(uint size, uint count, char *msg)
{
    uint i;

    if (map.verbose == 0)
	return;
    if (alloc_map == NULL) {
	alloc_map = calloc(1, 100*sizeof(memlist_t));
	map_size(alloc_map) = 100;
	first_free(alloc_map) = 1;
	overflow_list(alloc_map) = NULL;
    }
    if (size == 0) {	/* print... */
	int total = 0;
	if (!(map.verbose || count))
	    return;
	for (i=1; i < first_free(alloc_map); i++) {
	    total += alloc_map[i].size*alloc_map[i].count;
	    logmsg(LOGMEM, "-- used[%2d] size %3d count %d\n",
		i, alloc_map[i].size, alloc_map[i].count);
	}
	logmsg(LOGMEM, "-- TOTAL %d\n", total);
	return;
    }
    for (i=1; i < first_free(alloc_map); i++)
	if (alloc_map[i].size == size) {
	    if (count > 0) /* this is a free! */
		alloc_map[i].count -= count;
	    else
		alloc_map[i].count++;
	    return;
	}
    if (count > 0)
	panic("size not found in free %d\n", size);
    if (i >= map_size(alloc_map))
	panic("too many different sizes\n");
    logmsg(LOGMEM, "++ first alloc for size %d <%s>\n", size, msg);
    alloc_map[i].size = size;
    alloc_map[i].count = 1;
    first_free(alloc_map)++;
}
#endif

void
dump_alloc()
{
    record_alloc(0,0, NULL);
}

/*
 * Insert block x into the map m
 */
static void
mem_insert(memlist_t *m, mem_block_t *x)
{
    uint i, lim;

    if (m == NULL)		/* use the main map */
	m = mem_map;
    CHECK_PTR(x, "mem_insert");
    logmsg(V_LOGMEM, "mem_insert %p size %8d, map %p %2d/%8d\n",
	x, x->size, m, first_free(m), map_size(m));
    for (lim = first_free(m), i=1; i<lim; i++)
	if (m[i].size == x->size)	/* found the slot */
	    break;
    if (i >= map_size(m)) {		/* We ran out of slots		*/
	if (m != mem_map) {		/* Cannot realloc, must use slot 0 */
	    logmsg(LOGWARN, "Warning, must use slot 0 for map %p\n", m);
	    x->next = overflow_list(m);
	    overflow_list(m) = x;
	    return;
	} else {			/* Good, can reallocate the map */
	    if (map_size(m) > 128)
		logmsg(LOGWARN, "*** Warning, large map size %d\n", m[0].size);
	    map_size(m) *= 2;	/* Use twice the original size */
	    m = realloc(m, map_size(m)*sizeof(*m));
	    logmsg(LOGMEM, "mem_insert: realloc map with size %d\n",
		    map_size(m));
	    if (m == NULL)
		panic("failed map reallocation\n");
	    mem_map = m;
	}
    }
    if (i > first_free(m))
	panic("*** invalid mem_insert index %d should be %d max %d\n",
		i, first_free(m), map_size(m));
    if (i == first_free(m)) {		/* Previously empty slot */
	first_free(m)++;
	m[i].next = NULL;
	m[i].size = x->size;
	m[i].count = 0;
    }
    x->next =m[i].next;
    m[i].count++;
    m[i].next = x;
}

/*
 * merge map m into dst.
 * We expect the number of slots to be small, so rather than use
 * a complex sorting/merging algorithm, just concatenate the maps
 * unless there is major trouble which requires individual entries
 * to be merged in.
 */
int
mem_merge_maps(memlist_t *dst, memlist_t *m)
{
    uint have, need;
    uint i, j;
    uint saved = 0;	/* memory recovered */
    uint orig_src = first_free(m);

    CHECK_PTR1(m, "mem_merge_maps");
    if (dst == NULL)
	dst = mem_map;
    need = first_free(m) - 1;		/* slots we need */
    have = map_size(dst) - first_free(dst);	/* slots we have */
    if (have < need && dst == mem_map) {	/* can reallocate, do it */
	map_size(dst) += need - have;
	dst = realloc(dst, map_size(dst)*sizeof(memlist_t));
	logmsg(V_LOGMEM, "---- mem_merge_maps: realloc map with size %d\n",
		map_size(dst));
	if (dst == NULL)
	    panic("failed map reallocation\n");
	mem_map = dst;
	have = need;
    }

#if 0 	/* XXX debugging info */
    for (i=1; i < first_free(dst); i++)
	logmsg(V_LOGMEM, "-- dst[%2d] size %3d count %d\n",
		i, dst[i].size, dst[i].count);
    for (i=1; i < first_free(m); i++)
	logmsg(V_LOGMEM, "-- src[%2d] size %3d count %d\n",
		i, m[i].size, m[i].count);
#endif

    if (need <= have) {			/* easy... */
	for (i=1, j=first_free(dst); i < first_free(m) ; i++, j++) {
	    dst[j] = m[i];
	    saved += dst[j].size * dst[j].count;
	    if (dst == mem_map)
		record_alloc(dst[j].size, dst[j].count, "merge_map");
	}
	first_free(dst) = j;
	first_free(m) = 1;
    }
    j = first_free(m);			/* leftover lists */
    if (j > 1 || overflow_list(m) != NULL)
	logmsg(LOGWARN, "*** WARNING, full map (j %d, next %p)\n",
		j, overflow_list(m));

    /* copy individual elements from leftover lists if any */
    for (i=0; i < j; i++) {
	mem_block_t *x, *y;
	for (x = m[i].next; x; ) {
	    saved += x->size;
	    if (dst == mem_map)
		record_alloc(x->size, 1, "merge leftover");
	    y = x;
	    x = x->next;
	    mem_insert(dst, y);
	}
    }
    if (dst == mem_map)
	memory_in_use -= saved;
    logmsg(LOGMEM, "-- mem_merge_maps %d/%d, %d/%d freed %d in_use %d peak %d\n",
	first_free(dst), map_size(dst),
	orig_src, map_size(m),
	saved, memory_in_use, peak);
    dump_alloc();
    return saved;
}

void *
new_mem(memlist_t *m, uint size, __unused char *msg)
{
    uint i, cand;
    mem_block_t *x;

    if (m == NULL)
	m = mem_map;
    else
	panic("alloc from unsupported map XXX fixme\n");
    if (m == NULL)
	panic("memory not initialized\n");

    logmsg(V_LOGMEM, "-- new_mem %4d in_use %6d/%6d (map %p %2d/%2d)\n",
	size, memory_in_use, peak, m, first_free(m), map_size(m));
    for (cand = 0, i=1; i<first_free(m);) {
	logmsg(V_LOGMEM, "   slot %2d size %8d count %3d p %p\n",
		i, m[i].size, m[i].count, m[i].next);
	if (m[i].next == NULL) {	/* empty slot, grab the last one */
	    first_free(m)--;
	    m[i] = m[first_free(m)];
	    continue;
	}
	if (m[i].size == size) {		/* found it */
	    cand = i;
	    break;
	}
	if (m[i].size > size) {		/* potential candidate */
	    if (cand == 0 || m[i].size < m[cand].size)
		cand = i;
	}
	i++;
    }
    if (cand == 0) {
        logmsg(LOGWARN, "-- new_mem %4d in_use %6d/%6d (map %p %2d/%2d)\n",
		size, memory_in_use, peak, m, first_free(m), map_size(m));
	usleep(1000);	/* XXX give other processes a chance! */
	logmsg(V_LOGMEM, "sorry out of memory for %d bytes (%s)!\n", size, msg);
	for (i=1; i < first_free(m);i++)
	    logmsg(V_LOGMEM, "   slot %2d size %8d count %3d p %p\n",
		i, m[i].size, m[i].count, m[i].next);
	return NULL;
    }
    logmsg(V_LOGMEM, "-- new_mem need %d have %d at slot %d\n", 
	size, m[cand].size, cand);
    x = m[cand].next;
    CHECK_PTR(x, "new_mem-1");
    m[cand].count--;
    m[cand].next = x->next;
    x->next = NULL;
    if (m[cand].next == NULL) {	/* slot now empty, replace with last one */
	first_free(m)--;
	m[cand] = m[first_free(m)];
    } else if (cand != 1) {	/* move to front. */
	memlist_t y = m[cand];
	m[cand] = m[1];
	m[1] = y;
    }
    if (x->size - size > MIN_SIZE + sizeof(mem_block_t)) {
	/* split the block */
	mem_block_t *y = (mem_block_t *)(x->data + size);

	y->size = x->size - size - sizeof(mem_block_t);
	y->_magic = MY_MAGIC;
	x->size = size;
	mem_insert(m, y);
    }
    if (m == mem_map) {
	memory_in_use += x->size;
	if (memory_in_use > peak)
	    peak = memory_in_use;
    }
    record_alloc(x->size, 0, msg);
    x->_magic = MY_MAGIC_IN_USE;
    bzero(x->data, size);	/* XXX check this. */
    return x->data;
}

/*
 * Frees the block of memory p into pool m.
 * If a non-zero size is specified, the size is also checked with the
 * allocation, and the block is zeroed to detect stale pointers etc
 * (XXX the zeroing will go away once the code is debugged).
 */
void
mfree_mem(memlist_t *m, void *p, uint size)	/* equivalent to free(p) */
{
    mem_block_t *x = (mem_block_t *)p - 1;

    CHECK_PTR(x, "mfree_mem");
    if (x->_magic != MY_MAGIC_IN_USE)
	panic("block not in use, %p magic %p\n", x, x->_magic);
    x->_magic = MY_MAGIC;
    if (size != 0) {
	if (size > x->size)
	    panic("mfree_mem %p size %d real_size %d\n", p, size, x->size);
	bzero(p, size);
    }
    logmsg(V_LOGMEM, "mfree_mem %p size %d real_size %d\n", p, size, x->size);
    if (m == mem_map) {
	record_alloc(x->size, 1, "mfree_mem");
	memory_in_use -= x->size;
    }
    mem_insert(m, x);
}


/*
 * -- memory_clear
 *
 * Used by clients to reset mapping 
 *
 */
void 
memory_clear() 
{
    logmsg(LOGMEM, "freeing memory map\n");
    free(mem_map);
    mem_map = 0;
    return;
}

void
memory_init(uint chunk)
{
    static int done;
    mem_block_t *m;

    if (done)
	panic("*** can only do memory_init() once\n");
    done = 1;
    m = mmap((void *)0, chunk*1024*1024, PROT_WRITE|PROT_READ,
	    MAP_ANON|MAP_NOSYNC|MAP_SHARED, -1 /* fd */, (off_t)0);
    if (m == MAP_FAILED)
	panic("*** failed to allocate %d MB of mapped memory\n", chunk);
    logmsg(LOGMEM, "allocated %d MB of mapped memory\n", chunk);
    shared_low = shared_high = (char *)m;
    shared_high += chunk*1024*1024;

    m->size = chunk*1024*1024 - sizeof(mem_block_t);
    m->_magic = MY_MAGIC;
    m->next = NULL;

    /* allocate an initial, pseudo-empty map */
    mem_map = calloc(1, sizeof(memlist_t));
    map_size(mem_map) = 1;
    first_free(mem_map) = 1;
    overflow_list(mem_map) = NULL;
    mem_insert(NULL, m);
}

memlist_t *
new_memlist(uint entries)
{
    memlist_t *x = new_mem(NULL, sizeof(memlist_t)*(1+entries), "new_memlist_t" );

    map_size(x) = 1+entries;
    first_free(x) = 1;
    overflow_list(x) = NULL;
    return x;
}
