/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: query-comms.c,v 1.7 2005/01/22 11:27:14 dvercell Exp $
 */

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>		/* write, read */


#include "como.h"
#include "query.h"

/*
 * Communication support for the query modules.
 * Essentially pack/unpack messages and make sure that we read them
 * atomically even if they are sent over a TCP socket.
 */

/*
 * -- qrysend
 *
 * packages a message and sends it using socket sd
 *
 */
int
qrysend(int sd, uint16_t type, uint16_t code, char * msg, size_t len)
{
    char mybuf[4096]; 
    char *buf;
    qctl_t *ctl;
    qdata_t *data;
    size_t sz, wr; 

    bzero(mybuf, sizeof(mybuf));
    buf = mybuf;

    switch (type) { 
    case Q_ERR: 
	ctl = (qctl_t *) buf;
	ctl->type = type; 
	ctl->code = htons(code); 
	sz = sizeof(qctl_t); 
	break; 

    case Q_DATA: 
	data = (qdata_t *) buf;
	data->type = type; 
	sz = len + sizeof(qdata_t); 
	if (msg == NULL || sz > sizeof(mybuf))
	    return -1; 
	data->len = htons((uint16_t) sz); 
	bcopy(msg, data->rec, len); 
	break;

    case Q_ASCII: 
	buf = msg;
	sz = len;
	break;

    default: 
	return 0; 
    }

    wr = 0; 
    while (wr < sz) { 
	int ret; 
        ret = write(sd, buf+wr, sz - wr); 
        if (ret < 0) 
            return 0; 
	wr += ret; 
    } 

    return 1; 
}

/**
 * query_parse()
 * takes an http-like request and formats a message in the
 * standard format
 */
static qreq_t *
query_parse(char *buf, __unused int len)
{
    static qreq_t q;
    int i;
    char *p;

    bzero(&q, sizeof(q));
    q.type = Q_ASCII;
    q.len = sizeof(q);

    p = strstr(buf, "Module: ");
    if (p == NULL)
	goto error;
    for (i = 0, p = p+8; i < MODULENAME_MAX && p[i] > ' ' ; i++)
	q.module[i] = p[i];

    p = strstr(buf, "Filter: ");
    if (p == NULL) /* use a default value */
	strcpy(q.filter, "ALL");
    else {
	for (i = 0, p = p+8; i < FILTERSTR_MAX && p[i] >= ' ' ; i++)
	    q.filter[i] = p[i];
    }

    p = strstr(buf, "Tmin: ");
    if (p == NULL) {	/* use a default value */
	struct timeval t;
	gettimeofday(&t, NULL);
	q.start = t.tv_sec - 50;
    } else
	q.start = atoi(p+6);

    p = strstr(buf, "Tmax: ");
    if (p == NULL) {		/* use a default value */
	struct timeval t;
	gettimeofday(&t, NULL);
	q.end = t.tv_sec + 20;
    } else
	q.end = atoi(p+6);
    return &q;

error:
    logmsg(LOGQUERY, "Missing argument in %s\n", buf);
    return NULL;
}


/* 
 * -- qryrecv
 * 
 * receives a message from a TCP connection making sure that
 * the message is complete. 
 *
 */
qreq_t * 
qryrecv(int sd) 
{
    static char buf[4096]; 
    int rd; 
    qreq_t *req;
    size_t ofs, len; 

    req = (qreq_t *)buf;

    /*
     * Read as much as possible from the socket.
     * We accept two types of messages - one is the historical one, qreq_t,
     * the other one is an http-style string ending with a pair of \n
     */
    ofs = 0; 
    bzero(buf, sizeof(buf));
    for (;;) {
	rd = read(sd, buf + ofs, sizeof(buf) - ofs);
	logmsg(LOGSTORAGE, "qryrecv read returns %d\n", rd);
	if (rd <= 0)	/* other end closed connection ? */
	    return NULL; 
	ofs += rd;
	if (strstr(buf, "\n\n") != NULL ||
		strstr(buf, "\n\r\n") != NULL) /* found terminator */
	    return query_parse(buf, ofs);
	if (ofs > 0 && buf[0] < Q_MAX)	/* traditional string */
	    break;
    } 
    /* only accept queries, not replies */
    if (req->type != Q_QUERY)
	return NULL;
	
    /*  XXX this code is broken!!
     * first, try to read a number of bytes equal to the 
     * minimum request length. 
     */
    len = sizeof(qreq_t); 
    while (ofs < len) { 
	rd = read(sd, &buf[ofs], sizeof(buf) - ofs); 
	if (rd < 0) 
	    return NULL;
	if (rd == 0) { 	
	    /* connection closed gracefully (?!?!).
	     * consider it as an error 
	     */
	    return NULL; 
	}
	ofs += rd; 

	/* deal with optional arguments if any updating 
	 * the target message length 
	 */
	if (ofs >= sizeof(qreq_t))
	    len = (size_t) ntohs(req->len);
    }

    return req;
}
