/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: query-ondemand.c,v 1.17 2005/01/25 19:15:21 gianluca Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h> /* socket, bind, listen, accept */
#include <netinet/in.h> /* struct sockaddr_in */
#include <string.h>     /* bzero */
#include <errno.h>      /* errno */

#include "como.h"
#include "storage.h"
#include "query.h"

/* global state */
extern struct _como map;

/*
 * This code implements the body of the query-ondemand process.
 * One instance is forked by QUERY right after accepting the
 * connection, and this process only serves a single connection
 * then terminates.
 */

/*
 * -- getrecord
 *
 * This function reads a chunk from the file (max_recsize bytes long). 
 * It then passes it to the load() callback to get the timestamp
 * and the actual record length. 
 * Return values:
 * - on success, returns a pointer to the record, its length and its timestamp. 
 *   (with length > 0)
 * - on 'end of bytestream', returns NULL and len=0
 * - on 'error' (e.g. csmap failure), returns NULL and len != 0
 * - on 'lost sync' (bogus data at the end of a file), returns
 *   a non-null (but invalid!) pointer GR_LOSTSYNC
 *
 * XXX needs attention here, in case we lose sync and must move to the
 * next file.
 */
#define	GR_LOSTSYNC	((void *)getrecord)
static void *
getrecord(int fd, load_fn *ld, size_t *len, timestamp_t *ts)
{
    // XXX why these static vars here ???
    static int firstcall = 1; 
    static off_t ofs;
    char * ptr; 

    bzero(ts, sizeof(*ts));
    if (firstcall) { 
	ofs = csgetofs(fd); 
  	firstcall = 0; 
    } 

    /* 
     * mmap len bytes starting from last ofs. 
     * len bytes are supposed to guarantee to contain
     * at least one record to make sure the load() doesn't
     * fail (note that only load knows the record length). 
     */ 
    ptr = csmap(fd, ofs, len); 
    if (ptr == NULL) 
	return NULL;

    /* give the record to load() */
    *len = ld(ptr, *len, ts); 
    if (*len == 0) {
	/*
	 * We have lost sync. The only escape seems to be to seek
	 * to the beginning of the next file.
	 */
 	logmsg(LOGQUERY, "lost sync with bytestream move to next file\n");
	ofs = csseek(fd);
	ptr = GR_LOSTSYNC;
    } else {
	ofs += *len;
    }
    return ptr;
}


/*
 * -- query_ondemand
 *
 * This function is used for on-demand queries. It is called by
 * query_mainloop() and runs in a new process. It is in charge of
 * authenticating the query, finding the relevant module output
 * data and send them back to the requester. 
 * 
 * A query comes over a TCP socket with the following information: 
 *
 *  . name, the module to run (the shared object must exist already)
 *  . filter, filter expression
 *  . start, start timestamp
 *  . end, end timestamp
 * 
 * XXX as of now, query_ondemand requires that the module has been running
 *     during the interval of interest. this way it just has to find the 
 *     output file, read them and send them "as is" to the client. 
 *
 */
void
query_ondemand(int client_fd)
{
    module_t *mdl;
    qreq_t *req;
    int storage_fd, file_fd;
    enum q_type_t reply_fmt = Q_DATA;

    /* set the name of this process */
    map.procname = "qd"; 
    setproctitle("ONDEMAND");
    if (map.debug) {
	if (strstr(map.debug, map.procname) != NULL) {
	    logmsg(LOGWARN, "waiting 60s for the debugger to attach\n");
	    sleep(60);
	    logmsg(LOGWARN, "wakeup, ready to work\n");
	}
    }

    logmsg(LOGWARN, "starting query-ondemand #%d pid %d\n",
	client_fd, getpid()); 

    req = (qreq_t *)qryrecv(client_fd); 
    if (req == NULL) 
	panic("*** query-ondemand receiving query: %s\n", strerror(errno)); 
    /*
     * req->type == Q_QUERY for the traditional record-based query,
     *		Q_ASCII for the ascii-style one.
     * This is used to differentiate replies.
     */
    if (req->type == Q_ASCII)
	reply_fmt = Q_ASCII;

    logmsg(LOGQUERY,
        "got query (%d bytes); mdl: %s filter: %s\n",  
        ntohs(req->len), req->module, req->filter); 
    logmsg(0, "    from %d to %d\n", req->start, req->end); 

    /* 
     * check if the module is running using the same filter 
     * 
     * XXX right now we just check if the module exists and is using the 
     *     exactly same filter. in the future we will have to check 
     *     if the module has been running during the interval of interest. 
     *     if not, we have to run it on the stored packet trace. 
     *     furthermore, the filter should be semantically equivalent but
     *     right now we check the syntax as well (i.e., right now "A and B" 
     *     is not the same as "B and A"). 
     * 
     */
    for (mdl = map.modules; mdl; mdl = mdl->next) {
	 /* check module name */
	 if (strncmp(req->module, mdl->name, strlen(mdl->name)))
	     continue; 

	 /* check filter string */
	 if (!strncmp(req->filter, mdl->filter, strlen(mdl->filter))) 
	     break; 	/* found! */
    } 

    if (mdl == NULL) { 
	/* 
	 * no module found. return an error message 
	 * to the client. 
	 */
	logmsg(LOGWARN, "query module not found (%s)\n", req->module); 
	qrysend(client_fd, Q_ERR, ENOENT, NULL, 0); 
	close(client_fd);
	return; 
    }

    /* 
     * Check if query is interested in capture or export data. 
     * Make sure the load() callback is defined, open 
     * the module output file and then start reading the file 
     * and send the data back 
     */
    if (mdl->callbacks.load == NULL) 
	panic("callback load() doesn't exist\n");

    /*
     * connect to the storage process
     */
    storage_fd = create_socket("storage.sock", 0 /* client */);

    /* 
     * open the output file for reading.  
     */
    logmsg(LOGQUERY, "opening file for reading (%s)\n", mdl->output); 
    file_fd = csopen(mdl->output, CS_READER, 0, storage_fd); 
    if (file_fd < 0) 
	panic("opening file %s: %s", mdl->output, strerror(errno)); 

    /*  
     * now look for the start time in the file 
     * 
     * XXX we do this without seeking the file. we read all the
     *     records one by one to find the beginning. very inefficient.
     *     one day STORAGE will support timestamp-based seeks.
     *  
     */
    for (;;) { 
	uint t;
	timestamp_t ts;
	char * ptr; 
	ssize_t len;

        len = mdl->bsize; 
        ptr = getrecord(file_fd, mdl->callbacks.load, &len, &ts);
        if (ptr == NULL) {	/* no data, but why ? */
	    if (len == -1) 
		panic("reading from file %s\n", mdl->output); 

	    if (len == 0) {
		logmsg(LOGQUERY, "reached end of file %s\n", mdl->output); 
		break;
	    }
	}
	/*
	 * Now we have either good data or or GR_LOSTSYNC.
	 * We should not bother to differentiate because getrecord()
	 * clears the timestamp and so we remain in the loop anyways,
	 * but we do just for safety.
	 */
	if (ptr == GR_LOSTSYNC) {
	    logmsg(LOGQUERY, "lost sync, trying next file %s\n", mdl->output); 
	    continue;
	}
    	t = TS2SEC(ts);
    	if (t < req->start)	/* before the required time. */
	    continue;
    	if (t >= req->end) {
	    logmsg(LOGQUERY, "query completed\n"); 
	    break;
	}
	if (reply_fmt == Q_ASCII && mdl->callbacks.print != NULL) {
	    /*
	     * invoke the print() method to print data.
	     * Replace the final NULL with a \n if not there
	     */
	    ptr = mdl->callbacks.print(ptr);
	    len = strlen(ptr);
	    if (len > 0 && ptr[len-1] != '\n') {
	        ptr[len] = '\n';
		len++;
	    }
	}
	if (!qrysend(client_fd, reply_fmt, 0, ptr, len))
	    panic("sending data to the client: %s\n", strerror(errno));  
#if 0
	logmsg(LOGQUERY, "data record sent (%d bytes)\n", len); 
	logmsg(0, "--- %#x %#x %#x...\n", *((int32_t*)ptr), 
		*((int32_t*)ptr+4), *((int32_t*)ptr+8)); 
#endif
    }

    /* close the socket and the file */
    close(client_fd);
    csclose(file_fd); 
}
