/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: query.c,v 1.40 2005/01/23 19:52:20 gianluca Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h> /* socket, bind, listen, accept */
#include <netinet/in.h> /* struct sockaddr_in */
#include <string.h>     /* bzero */
#include <errno.h>      /* errno */

#include "como.h"
#include "storage.h"
#include "query.h"


/* 
 * 
 
 --- QUERY SUPPORT ---

CoMo supports three types of queries: static, on-demand and ad-hoc. 

Static queries are defined in the configuration file with the "sendto" 
keywords. Once the configuration file is parsed (in config.c) a process
is forked (como-qsendto) that handles all the static queries together. The
process periodically tries to open a TCP connection with the destinations. 
If the connection is setup then it sends all the output data of the related
module. If the connection drops, qsendto tries to resume and send the data 
exactly from where it stopped. 

On-demand queries are initiated by the remote client. They are handled by
a different process on the connection used to initiate the query. If the 
connection is lost, the query is stopped. If the module of the query has 
been running, como-ondemand just read the output file for the requested
time window. Otherwise it goes to the packet trace and runs the module 
internally (i.e. filter, callbacks, etc.). Then, it sends the output data. 

Ad-hoc queries are handled off-line and should revert ot on-demand queries
for which the module for sure didn't exist. 

 *
 */


/* global state */
extern struct _como map;


/* 
 * -- initserver
 * 
 * Initializes TCP socket to receive queries from clients.
 * It returns socket. 
 */
int
initserver()
{
    struct sockaddr_in saddr;
    int sd; 
    int opt;
    int ret; 

    sd = socket(AF_INET, SOCK_STREAM, 0);
    if (sd < 0)
        panic("*** creating the socket: %s\n", strerror(errno));

    /* allow local address reuse in TIME_WAIT */
    opt = 1;
    setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
   
    /* assign address to the socket */
    bzero(&saddr, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(QSERV_PORT); 
    saddr.sin_addr.s_addr = INADDR_ANY;
    ret = bind(sd, (struct sockaddr *)&saddr, sizeof(saddr));
    if (ret < 0)
        panic("*** binding address: %s\n", strerror(errno));

    listen(sd, SOMAXCONN);
    return sd; 
}


/*
 * -- query_mainloop
 *
 * It first forks for the qsendto process that will deal with 
 * all static queries. Then it waits for incoming connections and
 * on-demand, ad-hoc queries. 
 * 
 * The input parameters are the socket to communicate with the 
 * supervisor process and the pool of descriptors to communicate with 
 * the storage process. The pool is dimensioned to be equal to the 
 * maximum number of concurrent queries the system should handle plus 
 * the descriptor for the static queries.  
 *
 */
void
query_mainloop(__unused int fd) 
{
    int client_fd;
    pid_t pid;
    fd_set rx;
    int max_fd; 

#if 0 
    /* XXX we do not have query-static yet!
     *     need to define interface, behavior, etc.
     *     it should be used for automatic alert systems.
     */

    /* start the -como-query-static process */
    pid = fork(); 
    if (pid < 0)  
	logmsg(LOGWARN, "starting query-static: %s\n", strerror(errno)); 

    if (pid == 0) {  		
	/* this is query-static */
        map.procname = "qs";

	/* XXX query-static doesnt exist yet!! */
        /* query_static(); */

        exit(EXIT_FAILURE);
    }
#endif

    /* 
     * now start listening for query requests and fork new 
     * processes to deal with them 
     */
    client_fd = initserver(); 
    
    /* 
     * query will wait on a select for new queries or 
     * messages from the supervisor
     */
    max_fd = map.supervisor_fd > client_fd ? map.supervisor_fd : client_fd;
    FD_ZERO(&rx);
    FD_SET(client_fd, &rx);
    FD_SET(map.supervisor_fd, &rx);

    /* 
     * query loop. wait for incoming connections (from 
     * como_client applications). then fork a process to 
     * serve the request. 
     */ 
    for (;;) { 
        fd_set r;
        int ret;

        logmsg(V_LOGQUERY, "waiting for queries from clients\n"); 

	r = rx; 
	errno = 0;	/* clear previous errors */
        ret = select(max_fd + 1, &r, NULL, NULL, NULL);
	if (ret < 0) {
	    if (errno == EINTR)	/* someone interrupted us */
		continue;
	    logmsg(LOGWARN, "*** select failed: %s", strerror(errno));
	    continue;
	}

	if (FD_ISSET(client_fd, &r)) { 
            struct sockaddr_in addr;
            socklen_t len;
	    int cd; 
	    
            len = sizeof(addr);
            cd = accept(client_fd, (struct sockaddr *) &addr, &len); 
            if (cd < 0) {
                /* check if accept was unblocked by a signal */
                if (errno == EINTR)
                    continue;

                logmsg(LOGWARN, "accepting connection: %s\n", strerror(errno));
            }

  	    logmsg(LOGQUERY, "query from %s\n", inet_ntoa(addr.sin_addr)); 

	    /* fork a process to serve the query */
            pid = fork(); 	
	    if (pid < 0) 
		logmsg(LOGWARN, "fork query-ondemand: %s\n", strerror(errno));

            if (pid == 0) {	/* here is the child... */
		close(map.supervisor_fd);
		map.supervisor_fd = create_socket("supervisor.sock", 0);

                query_ondemand(cd); 
                close(cd);
                exit(EXIT_SUCCESS); 
            } else	/* parent */
		close(cd);
        }

        if (FD_ISSET(map.supervisor_fd, &r)) { 
	    /* XXX nothing implemented yet */
	} 
    }

    return; 	/* never reached */
}
