/*-
 * Copyright (c) 2004, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sniffer-bpf.c,v 1.22 2005/01/23 19:52:20 gianluca Exp $
 */


#include "sniffers.h"
#include "como.h"

#if !defined(linux)	/* no bpf on linux, use libpcap instead */
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <pcap.h>
#include <net/if.h>

#define MTU 1500
#define BUFSIZE (100 * (sizeof(struct bpf_hdr) + MTU))

/*
 * Sniffer that accesses directly the bpf device.
 * Note that because of the use of static variables to hold
 * sniffer's state here, we can only support one instance of this
 * sniffer.
 */
static int type;       /* indicate link layer type */
static int chkdir;     /* indicate if traffic direction has to be checked */

/*
 * sniffer_open
 */
static int
sniffer_start(char *ifname, int dir)
{
    char device[sizeof("/dev/bpf000")];
    int bufsize = BUFSIZE;
    struct ifreq ifr;
    int fd = -1;
    int n;

    if (strlen(ifname)+1 > sizeof(ifr.ifr_name)) {
        errno = EINVAL;
        return -1;
    }

    n = 0;
    do {
        sprintf(device, "/dev/bpf%d", n++);
        fd = open(device, O_RDWR);
    } while (fd < 0 && errno == EBUSY && n < 1000);

    if (fd < 0) {
	logmsg(LOGWARN, "sniffer-bpf: cannot find a usable bpf device\n");
        return -1;
    }

    if (ioctl(fd, BIOCSBLEN, &bufsize)) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCSBLEN failed\n");
        close(fd);
        return -1;
    }

    n = 1;
    if (ioctl(fd, BIOCIMMEDIATE, &n) < 0) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCIMMEDIATE failed\n");
        close(fd);
        return -1;
    }

    n = 0;
    if (ioctl(fd, BIOCSRTIMEOUT, &n) < 0) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCSRTIMEOUT failed\n");
        close(fd);
        return -1;
    }

    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    if (ioctl(fd, BIOCSETIF, &ifr)) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCSETIF failed\n");
        close(fd);
        return -1;
    }

    if (ioctl(fd, BIOCGDLT, &n) < 0) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCGDLT failed\n");
        close(fd);
        return -1;
    }

    if (ioctl(fd, BIOCPROMISC) < 0) {
	logmsg(LOGWARN, "sniffer-bpf: BIOCPROMISC failed\n");
        close(fd);
        return -1;
    }

    chkdir = 0;
    switch (n) {
    case DLT_EN10MB:
	type = BPF_ETH;
	chkdir = dir;
	break;

    default:
	type = -1;  /* unknown type */
	break;
    }

    return fd;
}


/*
 * sniffer_next
 *
 * Fill a structure with a copy of the next packet and its metadata.
 * Return 1 if packet is available, 0 if not and we need to read again.
 *
 * BPF: each packet is preceded by the following header:
 *  struct timeval  bh_tstamp;      time stamp
 *  bpf_u_int32     bh_caplen;      length of captured portion
 *  bpf_u_int32     bh_datalen;     original length of packet
 *  u_short         bh_hdrlen;      length of bpf header (this struct
 *                                      plus alignment padding)
 */
static int
sniffer_next(int fd, pkt_t *hdr)
{
    static char buf[BUFSIZE];   /* base of the capture buffer */
    struct bpf_hdr *bh;         /* BPF header */
    char *base;                 /* where to copy from */
    int len   = 0;              /* remaining bytes in buffer */
    int ofs   = 0;              /* next pkt to read in buffer */
    uint npkt = 0;
    uint hl;

    if ((len = read(fd, buf, BUFSIZE)) < 0)
        return -1;

    while (ofs < len)
    {
        bh = (struct bpf_hdr *)(buf + ofs);

        if (len < ofs + (int)sizeof(struct bpf_hdr))
            return -1; /* header incomplete */

        if (len < ofs + (int)bh->bh_hdrlen + (int)bh->bh_caplen)
            return -1; /* packet incomplete */

        if (MAX_PKTS < (npkt+1) * sizeof(pkt_t))
            return -1; /* not enough space in output buffer */

        /* ok data is good now */
        hdr[npkt].ts = TIME2TS(bh->bh_tstamp.tv_sec, bh->bh_tstamp.tv_usec);
        hdr[npkt].caplen = bh->bh_caplen;
        hdr[npkt].len = bh->bh_datalen;

        /* skip BPF header */
        base = buf + ofs + bh->bh_hdrlen;

        /* determine traffic direction, if applicable */
        if (chkdir)
            hdr[npkt].dir = get_direction((struct _como_machdr *)base);

        /* increment ofs to next packet */
        ofs += BPF_WORDALIGN(bh->bh_hdrlen + bh->bh_caplen);

        /* skip link layer header */
        base += skip_layertwo(type); 

        /* copy the header */
        hdr[npkt].ih = *(struct _como_iphdr *)base;

        /* copy the rest */
        hl = (hdr[npkt].ih.vhl & 0x0f)*4;   /* skip IP header */
        if (hdr[npkt].caplen >= hl)         /* copy next header */
        {
            base += hl;
            hl = hdr[npkt].len - hl;        /* leftover bytes */
            if (hl > sizeof(hdr[npkt].p))
                hl = sizeof(hdr[npkt].p);
            bcopy(base, &hdr[npkt].p, hl);
        }

        /* increment the number of processed packets */
        npkt++;
    }

    /* return the number of copied packets */
    return npkt;
}


/*
 * sniffer_close
 */
static void
sniffer_stop(int fd)
{
    close(fd);
}

struct _sniffer bpf_sniffer = {
	"bpf", sniffer_start, sniffer_next, sniffer_stop,
	0 /* flags */ };
#else	/* dummy entry when no bpf is available */
struct _sniffer bpf_sniffer = {
	"no-sniffer", NULL, NULL, NULL, 0 };
#endif
