/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sniffer-dag.c,v 1.11 2005/01/23 19:52:20 gianluca Exp $
 */

//TODO: Try to open another stream if 0 is used by another process??

#include "sniffers.h"
#include "como.h"

#ifdef USEDAG

#include <dagapi.h>
#include <string.h>     /* bcopy */


#define BUFSIZE   (4 * 1024 * 1024)

int chkdir; /* indicate if traffic direction has to be checked */


/*
 * edag_open
 */
static int
sniffer_start(char *ifname, int dir)
{
    int fd;
    struct timeval tout, poll;

    /* open DAG */
    if ((fd = dag_open(ifname)) < 0)
        return -1;  /* errno is set */

    /* configure DAG: full packet capture, varlen records */
    if (dag_configure(fd, "slen=1536 varlen") < 0)
        return -1;  /* errno is set */

    /* attach to stream 0 and set the extra window */
    if (dag_attach_stream(fd, 0, 0, BUFSIZE) < 0)
        return -1;  /* errno is set */

    /* start capture on stream 0 */
    if (dag_start_stream(fd, 0))
        return -1;  /* errno is set */

    /* init DAG polling parameters to return immediately */
    tout.tv_sec  = 0;
    tout.tv_usec = 0;   /* disable wait timeout */
    poll.tv_sec  = 0;
    poll.tv_usec = 0;   /* no sleep */
    if (dag_set_stream_poll(fd, 0, 0, &tout, &poll) < 0)
        return -1;      /* errno is set */

    /* indicate if traffic direction has to be checked */
    chkdir = dir;

    return fd;
}


/*
 * edag_next
 */
static int
sniffer_next(int fd, pkt_t *hdr)
{
    static void *bot = NULL;    /* pointer to bottom of stream buffer */
    void *top = NULL;           /* pointer to top of stream buffer */
    dag_record_t *rec;          /* pointer to current record */
    char *base;                 /* where to copy from */
    uint num  = 0;              /* number of processed bytes */
    uint npkt = 0;              /* number of pkts processed */
    int len;                    /* record length */
    uint hl;                    /* IP header length */

    /* read ERF records from stream 0 */
    if ((top = dag_advance_stream(fd, 0, &bot)) == NULL)
        return -1; /* errno is set */

    /* check if we read something */
    if (top - bot == 0)
        return 0;

    while ((top-bot > dag_record_size) &&  num+dag_record_size < BUFSIZE)
    {
        /* check if we have enough space in output buffer */
        if (MAX_PKTS < (npkt+1) * sizeof(pkt_t))
            return npkt; /* next call will return this remainder pkts too */

        /* access to packet record */
        rec = (dag_record_t *)bot;
        len = ntohs(rec->rlen);

        /* check if entire record is available */
        if (top - bot < len)
            return npkt;

        /* check if record would go over extra window */
        if (num + len > BUFSIZE)
            return npkt;

        /* ok data is good now */
        hdr[npkt].ts     = rec->ts;
        hdr[npkt].caplen = len;
        hdr[npkt].len    = ntohs(rec->wlen);

        /* determine traffic direction, if applicable */
        if (chkdir)
            hdr[npkt].dir = rec->flags.iface;

        bot += len; /* increment bottom pointer to next packet */
        num += len; /* update number of processed bytes */

        /* skip ERF + link layer header */
        base = bot + dag_record_size;
        base += skip_layertwo(rec->type); 

        /* copy the header */
        hdr[npkt].ih = *(struct _como_iphdr *)base;

        /* copy the rest */
        hl = (hdr[npkt].ih.vhl & 0x0f)*4;   /* skip IP header */
        if (hdr[npkt].caplen >= hl)         /* copy next header */
        {
            base += hl;
            hl = hdr[npkt].len - hl;        /* leftover bytes */
            if (hl > sizeof(hdr[npkt].p))
                hl = sizeof(hdr[npkt].p);
            bcopy(base, &hdr[npkt].p, hl);
        }

        /* increment the number of processed packets */
        npkt++;
    }

    /* return the number of copied packets */
    return npkt;
}


static void
sniffer_stop(int fd)
{
    /* stop capture on stream 0 */
    dag_stop_stream(fd, 0);

    /* detach from stream 0 */
    dag_detach_stream(fd, 0);

    /* close DAG */
    dag_close(fd);
}
sniffer_t dag_sniffer = { "dag",
	sniffer_start, sniffer_next, sniffer_stop, SNIFF_POLL };
#else
sniffer_t dag_sniffer = { "no-sniffer", NULL, NULL, NULL, 0 };
#endif
