/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sniffer-erf.c,v 1.11 2005/01/23 19:52:20 gianluca Exp $
 */

#include <fcntl.h>
#include "como.h"
#include "sniffers.h"

#ifdef USEDAG

#include <dagapi.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>     /* bcopy */


#define MTU 1500
#define BUFSIZE (100 * (sizeof(dag_record_size) + MTU))

static int chkdir; /* indicate if traffic direction has to be checked */


/*
 * erf_open
 */
static int
sniffer_start(char *file, int dir)
{
    int fd;

    /* open the ERF trace file */
    if ((fd = open(file, O_RDONLY)) < 0)
        return -1;  /* errno is set */

    /* indicate if traffic direction has to be checked */
    chkdir = dir;

    return fd;
}


/*
 * erf_next
 */
static int
sniffer_next(int fd, pkt_t *hdr)
{
    static char buf[BUFSIZE];   /* base of the capture buffer */
    static uint num  = 0;       /* remaining bytes in buffer */
    dag_record_t *rec;          /* pointer to current record */
    char *base;                 /* where to copy from */
    uint ofs  = 0;              /* next pkt to read in buffer */
    uint npkt = 0;              /* number of pkts processed */
    int len;                    /* record length */
    uint hl;                    /* IP header length */
    int i;                      /* number of bytes readed */

    /* read ERF records from fd */
    if (((i = read(fd, buf + num, BUFSIZE - num)) < 0) || (!i && !num))
        return -1;

    /* account for pending data to process */
    num += i;

    while (ofs < num)
    {
        /* check if we have enough space in output buffer */
        if (MAX_PKTS < (npkt+1) * sizeof(pkt_t))
            goto need_more;

        /* check if header is complete */
        if (num < ofs + sizeof(dag_record_t))
            goto need_more;

        /* access to packet record */
        rec = (dag_record_t *)(buf + ofs);
        len = ntohs(rec->rlen);

        /* check if entire record is available */
        if (num < ofs + len)
            goto need_more;

        /* ok data is good now */
        hdr[npkt].ts     = rec->ts;
        hdr[npkt].caplen = len;
        hdr[npkt].len    = ntohs(rec->wlen);

        /* determine traffic direction, if applicable */
        if (chkdir)
            hdr[npkt].dir = rec->flags.iface;

        /* increment ofs to next packet */
        ofs += len;

        /* skip ERF + link layer header */
        base = buf + ofs + dag_record_size + skip_layertwo(rec->type);

        /* copy the header */
        hdr[npkt].ih = *(struct _como_iphdr *)base;

        /* copy the rest */
        hl = (hdr[npkt].ih.vhl & 0x0f) * 4; /* skip IP header */
        if (hdr[npkt].caplen >= hl)         /* copy next header */
        {
            base += hl;
            hl = hdr[npkt].len - hl;        /* leftover bytes */
            if (hl > sizeof(hdr[npkt].p))
                hl = sizeof(hdr[npkt].p);
            bcopy(base, &hdr[npkt].p, hl);
        }

        /* increment the number of processed packets */
        npkt++;
    }

    /* all packets of buffer have been readed */
    num = 0;

    /* return the number of copied packets */
    return npkt;

need_more: /* we have a partial packet on the end of the buffer */
    num -= ofs;
    bcopy(buf + ofs, buf, num);

    return npkt;
}


static void
sniffer_stop(int fd)
{
    close(fd);
}

sniffer_t erf_sniffer = { "erf",
	sniffer_start, sniffer_next, sniffer_stop, 0 };
#else
sniffer_t erf_sniffer = { "no-sniffer", NULL, NULL, NULL, 0 };
#endif
