/*-
 * Copyright (c) 2004, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sniffer-libpcap.c,v 1.7 2005/01/23 19:52:20 gianluca Exp $
 */

#include <sys/stat.h>
#include <fcntl.h>      /* open */
#include <unistd.h>     /* close */
#include <pcap.h>
#include <string.h>     /* memset */

#include "sniffers.h"
#include "como.h"

static char errbuf[PCAP_ERRBUF_SIZE];

static int skip_val = 0;

/* 
 * The integer returned by libpcap_open and used throughout the
 * other functions is actually a fake file descriptor chosen so
 * that select(3)s are always successful. A real pcap handle is
 * associated internally to every such decriptor.
 */

typedef struct _sl_fdlist {
    struct _sl_fdlist *next;
    int fd;
    pcap_t *pcap;
    int dir;         /* indicates if traffic direction should be checked */
} sl_fdlist;

/* The list of open pcap_handles, each with its own fake file descriptor */
static sl_fdlist *fdlist = NULL;

/*
 * libpcap_open
 */
static int
sniffer_start(char *ifname, int dir)
{
    sl_fdlist *elem;
    pcap_t* pt;
    
    errbuf[0] = '\0';
    
    /* TODO: should we ask for promiscuous mode? */
    pt = pcap_open_live( ifname, 96, 1 /* promisc */,
		1 /* timeout ms */, errbuf );
    
    if ( pt == NULL ) {
        fprintf( stderr, "%s", errbuf );
        return -1;
    }
    if ( errbuf[0] != '\0' )
        fprintf( stderr, "%s", errbuf );
    
    switch( pcap_datalink(pt) ) {
    case DLT_NULL:
	skip_val = 4;
	break;
    case DLT_EN10MB:
	skip_val = 14;
	break;
    case DLT_RAW:
	skip_val = 0;
	break;
    case DLT_LINUX_SLL:
	skip_val = 16;
	break;
    default:
	/* If such an error occurs, chances are that adding a suitable
	 * case branch will do the job */
	fprintf( stderr, "libpcap sniffer: Unrecognized datalink format" );
	pcap_close(pt);
	return -1;
    }
    
    /* Then, we add a new item to the list of open pcap descriptors */
    elem = malloc( sizeof(sl_fdlist) );
    if (elem == NULL) {
        perror( "sniffer-libpcap: malloc" );
        pcap_close(pt);
        return -1;
    }
    
    elem->next = fdlist;
    elem->fd = pcap_fileno(pt);
    elem->pcap = pt;
    elem->dir = dir;
    
    fdlist = elem;
    
    /* Return a select-compatible file descriptor */
    return elem->fd;
}

/*
 * libpcap_next
 *
 * Reads one packet and fills in a pkt_t struct accordingly.
 * Returns how many packets were read or -1 if an unrecoverable
 * error occurred.
 *
 * The raw data format depends on the input device.
 * Using libpcap, each packet is preceded by the following header:
 *   struct timeval ts;    time stamp
 *   int32 caplen;         length of portion present
 *   int32 len;            length of this packet (off wire)
 */
static int
sniffer_next(int fd, pkt_t *hdr)
{
    pcap_t *pt = NULL;
    int chkdir = 0;
    sl_fdlist *p = NULL;
    
    struct pcap_pkthdr pkthdr;
    u_char *pkt = 0;
    uint npkt = 0;
    uint hl = 0;
    
    /* get the pcap handle associated to fd */
    for (p = fdlist; p != NULL; p = p->next)
        if (p->fd == fd) {
            pt = p->pcap;
            chkdir = p->dir;
            break;
        }
    
    if (!pt) {
        fprintf(stderr, "sniffer-libpcap: unknown descriptor");
        return -1;
    }
    
    memset( &pkthdr, 0, sizeof( struct pcap_pkthdr ) );
    
    /*
     * TODO: call pcap_dispatch directly
     * and retreive all the available packages
     */
    pkt = (u_char*)pcap_next( pt, &pkthdr );
    if (pkt == NULL)
        return 0;
    
    hdr[0].ts      = TIME2TS(pkthdr.ts.tv_sec, pkthdr.ts.tv_usec);
    hdr[0].caplen  = pkthdr.caplen;
    hdr[0].len     = pkthdr.len;
    hdr[0].n_frags = 0;                    /* XXX: provisional */
    if (chkdir)
        hdr[0].dir = get_direction((struct _como_machdr *)pkt);
    hdr[0].mach = *( struct _como_machdr* )pkt;
    pkt += skip_val;
    
    hdr[0].ih = *( struct _como_iphdr* )pkt;
    
    /* IPV4 packets start with a 4 bit version number (0x4)    *
     * and a 4 bit header word-length                          */
    hl = ( ( hdr[0].ih.vhl & 0x0f ) << 2 );
    
    /* TODO: Almost blindly copied from sniffer-pcap...        * 
     *       should be checked                                 *
     * TODO: When could caplen be less than hl? Not sure       */
    if (hdr[0].caplen >= hl) {             /* copy next header */
        pkt += hl;
        hl = hdr[0].len - hl;                /* leftover bytes */
        if (hl > sizeof(hdr[npkt].p))
            hl = sizeof(hdr[npkt].p);
        bcopy(pkt, &hdr[0].p, hl);
    }
    
    return 1;
}

/*
 * libpcap_close
 */
static void
sniffer_stop(int fd)
{
    sl_fdlist *p, **l;
    int found = 0;
    
    for (l = &fdlist, p = *l; p != NULL; l = &p->next, p = *l)
        if (p->fd == fd) {
            found = 1;
            close(p->fd);
            pcap_close(p->pcap);
            *l = p->next;
            free(p);
            break;
        }
    
    if (!found)
        fprintf(stderr, "sniffer-libpcap: unknown descriptor");
}

struct _sniffer libpcap_sniffer = { "libpcap",
    sniffer_start, sniffer_next, sniffer_stop, 0 };
