/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sniffer-pcap.c,v 1.14 2005/01/23 19:52:20 gianluca Exp $
 */

#include "como.h"
#include "sniffers.h"

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <pcap.h>	/* DLT_* on linux */
// #include <net/bpf.h>
#include <net/if.h>

#define MTU 1500
#define BUFSIZE (100 * (sizeof(pcap_hdr_t) + MTU))


typedef struct
{
    struct timeval ts;
    int caplen;
    int len;
} pcap_hdr_t;

static int type;       /* indicate link layer type */
static int chkdir;     /* indicate if traffic direction has to be checked */

static int
sniffer_start(char *file, int dir)
{
    uint32_t hdr[6];
    int i;
    int fd;

    fd = open(file, O_RDONLY);
    if (fd < 0)
        panic("failed to open file %s\n", file);
    i = read(fd, &hdr, sizeof(hdr));
    if (i != sizeof(hdr))
        panic("failed to read pcap header\n");

    chkdir = 0;
    switch (hdr[5])
    {
        case DLT_EN10MB:
            type = BPF_ETH;
            chkdir = dir;
            break;
        default:
            type = -1;  /* unknown type */
            break;
    }

    return fd;
}


/*
 * sniffer_next
 *
 * Fill a structure with a copy of the next packet and its metadata.
 * Return 1 if packet is available, 0 if not and we need to read again.
 *
 * The raw data format depends on the input device:
 * PCAP: each packet is preceded by the following header
 *	struct timeval ts;  time stamp
 *	int caplen;         length of portion present
 *	int len;            length this packet (off wire)
 *
 */
static int
sniffer_next(int fd, pkt_t *hdr)
{
    static char buf[BUFSIZE];   /* base of the capture buffer */
    static uint len = 0;        /* remaining bytes in buffer */
    pcap_hdr_t *ph;             /* PCAP header */
    char *base;                 /* where to copy from */
    uint ofs  = 0;              /* next pkt to read in buffer */
    uint npkt = 0;
    uint hl;
    int i;

    errno = 0;
    if (((i = read(fd, buf + len, BUFSIZE - len)) < 0) || (!i && !len))
        return -1;

    len += i;

    while (ofs < len)
    {
        ph = (pcap_hdr_t *)(buf + ofs);

        if (len < ofs + sizeof(pcap_hdr_t))
            goto need_more; /* header incomplete */

        if (len < ofs + sizeof(pcap_hdr_t) + ph->caplen)
            goto need_more; /* packet incomplete */

        if (MAX_PKTS < (npkt+1) * sizeof(pkt_t))
            goto need_more; /* not enough space in output batch */

        /* ok data is good now */
        hdr[npkt].ts = TIME2TS(ph->ts.tv_sec, ph->ts.tv_usec);
        hdr[npkt].caplen = ph->caplen;
        hdr[npkt].len = ph->len;

        /* skip PCAP header */
        base = buf + ofs + sizeof(pcap_hdr_t);

        /* determine traffic direction, if applicable */
        if (chkdir)
            hdr[npkt].dir = get_direction((struct _como_machdr *)base);

        /* increment ofs to next packet */
        ofs += sizeof(pcap_hdr_t) + ph->caplen;

        /* skip link layer header */
        base += skip_layertwo(type); 

        /* copy the header */
        hdr[npkt].ih = *(struct _como_iphdr *)base;

        /* copy the rest */
        hl = (hdr[npkt].ih.vhl & 0x0f) * 4; /* skip IP header */
        if (hdr[npkt].caplen >= hl)         /* copy next header */
        {
            base += hl;
            hl = hdr[npkt].len - hl;        /* leftover bytes */
            if (hl > sizeof(hdr[npkt].p))
                hl = sizeof(hdr[npkt].p);
            bcopy(base, &hdr[npkt].p, hl);
        }

        /* increment the number of processed packets */
        npkt++;
    }

    /* all packets of buffer have been readed */
    len = 0;

    /* return the number of copied packets */
    return npkt;

need_more:
    len -= ofs;
    bcopy(buf+ofs, buf, (uint)len);
    
    return npkt;
}


static void
sniffer_stop (int fd)
{
    close(fd);
}

sniffer_t pcap_sniffer = {
	"pcap", sniffer_start, sniffer_next, sniffer_stop,
	0 /* flags */ };
