/*-
 * Copyright (c) 2003 - Intel Corporation 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: storage-client.c,v 1.25 2005/01/25 19:15:21 gianluca Exp $
 *
 */

/*
 *
This file implements the client side for processes trying to
access files in CoMo. The interface to the storage module has
four methods, and provides an mmap-like interface.

  int csopen(const char * name, int mode, size_t size, int sd)

	name		is the file name
	mode		is the access mode, CS_READER or CS_WRITER
 	size		is the max bytestream size (CS_WRITER only)
	sd		is the (unix-domain) socket used to talk to the
			daemon supplying the service.

	Returns an integer to be used as a file descriptor,
	or -1 on error.

  off_t csgetofs(int fd)

	fd		is the file descriptor

	Returns the offset in the current file.

  void *csmap(int fd, off_t ofs, ssize_t * sz)

	fd		is the file descriptor
	ofs		is the offset that we want to map
	*sz		on input, the desired size;
			on output, the actual size;

	Flushes any unmapped block, and returns a new one.
	Returns a pointer to the mapped region.

  off_t csseek(int fd)

	fd		is the file descriptor

	Moves to the beginning of the next file, unmapping any
	mapped region.

  void csclose(int fd)  

	fd		is the file descriptor

	Flushes any unmapped block, and closes the descriptor.
  
 *
 */

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <assert.h>


#include "como.h"
#include "storage.h"

/* 
 * Client-side file descriptor 
 * 
 * This data structure contains the minimal information a client
 * needs to know about the open file, i.e., the id, the current offset, 
 * the address and size of the current mmapped block, the start offset
 * of the block and the socket descriptor to communicate with the storage
 * process.
 * 
 */
typedef struct { 
    int sd;			/* unix socket 
				 * XXX not that nice to have it here */ 
    char * name; 		/* file name, dynamically allocated */
    int mode;			/* file access mode */
    int fd; 			/* OS file descriptor */
    int id; 			/* STORAGE file ID */
    off_t off_file; 		/* start offset of current file */
    void * addr; 		/* address of memory mapped block */
    size_t size; 		/* size of memory mapped block */
    off_t offset; 		/* bytestream offset of the mapped block */
} csfile_t;

/* file descriptors for open files */
static csfile_t * files[CS_MAXCLIENTS];

/*
 * -- waitforack
 *
 * Wait to receive an acknowledgemnt to a request 
 * sent to the server. 
 *
 */
static void
waitforack(int sd, csmsg_t * m)
{
    fd_set rd;
    int ret = 0;
    struct timeval t;

again:
    FD_ZERO(&rd);
    FD_SET(sd, &rd);
    t.tv_sec = 5;
    t.tv_usec = 0;
    /* wait for the message */
    ret = select(sd + 1, &rd, NULL, NULL, &t);
    if (ret == 0) {
	logmsg(LOGWARN, "*** waitforack: TIMEOUT [%d] %x\n", sd, m);
	goto again;
    }
	
    if (ret < 0)
	panic("select error (%s)\n", strerror(errno));

    /* read the message */
    ret = read(sd, m, sizeof(csmsg_t));
    if (ret != sizeof(csmsg_t)) 
	panic("read error: got %d, expected %d\n", ret, sizeof(csmsg_t)); 

    return;
}

/* 
 * -- csopen
 * 
 * csopen() opens a file managed by the storage process. It 
 * uses the unix socket to communicate its intentions. The function 
 * returns an integer used as file descriptor, or -1 on error.
 * Obviously the descriptor is not select()-able.
 */ 
int
csopen(const char * name, int mode, size_t size, int sd) 
{ 
    csfile_t * cf;
    csmsg_t m;
    int fd; 

    if (mode != CS_READER && mode != CS_WRITER) {
	logmsg(LOGWARN, "csopen: [%s] bad mode %d\n", name, mode);
	return -1;
    }

    logmsg(LOGSTORAGE, "csopen: [%s] %s\n", name,
	mode == CS_READER ? "CS_READER" : "CS_WRITER" );

    /* look for an empty file descriptor */
    for (fd = 0; fd < CS_MAXCLIENTS && files[fd] != NULL; fd++)
	;

    if (fd == CS_MAXCLIENTS) {
	logmsg(LOGWARN, "Too many open files, cannot open %s\n", name); 
	return -1;
    } 

    /* 
     * file descriptor found, prepare and send the request 
     * message to the storage process
     */
    bzero(&m, sizeof(m));
    m.type = S_OPEN;
    m.id = 0;
    m.arg = mode;
    m.size = size;
    strncpy(m.name, name, FILENAME_MAX); 

    if (write(sd, (char *) &m, sizeof(m)) < 0) 
	panic("sending message to storage: %s\n", strerror(errno)); 

    /* wait for the acknowledgment */
    waitforack(sd, &m);

    if (m.type == S_ERROR) {
	logmsg(LOGWARN, "error opening file %s: %s\n", name, strerror(m.arg));
	return -1;
    }

    /* 
     * allocate a new file descriptor and initialize it with 
     * the information in the message. 
     */
    cf = safe_calloc(1, sizeof(csfile_t)); 
    cf->sd = sd;
    cf->name = strdup(name);
    cf->mode = mode;
    cf->id = m.id;
    cf->off_file = m.ofs;  
    cf->offset = m.ofs + m.size;  // XXX probably unused
    cf->fd = -1; 

    files[fd] = cf;
    return fd; 
}
    

/* 
 * -- csmap
 * 
 * this function provides an interface similar to mmap. 
 * it maps in memory a region of the file. before doing so, informs 
 * the server of this intentions and makes sure that it can do it. 
 *  
 * the acknowledgement from the server carries two pieces of information: 
 *   . offset, that is set to the start offset of the file that 
 *     contains the block.
 *   . size, that indicates the size of the region that can be memory mapped. 
 *     (different only if the region overlaps multiple files). 
 * 
 * this function returns a pointer to the memory mapped region and the size
 * of the region. 
 *
 * csmap1 is overloaded to handle both csmap and csseek
 */
static void * 
cs_do_everything_else(int fd, off_t ofs, ssize_t * sz, int method) 
{
    csfile_t * cf;
    csmsg_t out, in;
    int flags;
    int diff;

    logmsg(V_LOGEXPORT, "csmap: file descriptor %d\n", fd);
    if (fd < 0 || fd >= CS_MAXCLIENTS || files[fd] == NULL) {
	logmsg(LOGWARN, "Bad file descriptor %d\n", fd); 
	*sz = -1;
	return NULL;
    } 
    cf = files[fd];

    /* 
     * Package the request and send it to the server. 
     * 
     * XXX this could be optimized if the client has already mmapped 
     *     the same offset. Anyway, we will do all the optimization later.
     */
    bzero(&out, sizeof(out)); 
    out.id = cf->id;
    if (method == S_REGION) {
	out.type = method;
	out.size = *sz;
	out.ofs = ofs; /* the map offset */
    } else {	/* all the seek variants */
	out.type = S_SEEK;
	out.arg = CS_SEEK_FILE_NEXT;
    }

    /* send the request out */
    if (write(cf->sd, &out, sizeof(out)) < 0) {
	logmsg(LOGWARN, "message to storage: %s\n", strerror(errno)); 
	*sz = -1;
	return NULL;
    }

    /* block waiting for the acknowledgment */
    waitforack(cf->sd, &in);

    switch (in.type) { 
    case S_ERROR: 
	logmsg(LOGWARN, "csmap error in %s: %s\n", cf->name, strerror(in.arg));
	errno = in.arg;
	*sz = -1;
	return NULL;

    case S_ACK: 
	/* 
	 * The server acknowledged our request,
	 * unmap the current block (both for csmap and csseek)
	 */
	if (cf->addr)
	    munmap(cf->addr, cf->size);

	if (method == S_REGION) {
	    /* 
	     * A zero sized block indicates end-of-bytestream. If so, close
	     * the current file and return an EOF as well.
	     */
	    if (in.size == 0) { 
		close(cf->fd);
		*sz = 0;
		return NULL;
	    }
	} else {	/* seek variants */
	    /* the current file is not valid anymore */
	    close(cf->fd);
	    cf->off_file = in.ofs;
	    cf->fd = -1;
	    *sz = (size_t)in.ofs;	/* return value */
	    return NULL;		/* we are done */
	}
	break;

    default: 
	panic("unknown msg type %d from storage\n", in.type);
	break;
    }

    if ((ssize_t)in.size == -1)
	panic("unexpected return from storage process");

    /*
     * now check if the requested block is in the same file (i.e., 
     * the server replied with an offset that is the first offset of 
     * the currently open file. if not, close the current file and 
     * open a new one. 
     */
    if (cf->fd < 0 || in.ofs != cf->off_file) {
	char * nm;

	if (cf->fd >= 0)
	    close(cf->fd);

#ifdef linux
	flags = (cf->mode == CS_READER)? O_RDWR : O_RDWR|O_APPEND; 
#else
	flags = (cf->mode == CS_READER)? O_RDONLY : O_WRONLY|O_APPEND; 
#endif
	asprintf(&nm, "%s/%016llx", cf->name, in.ofs); 
	cf->fd = open(nm, flags, 0666);
	if (cf->fd < 0)
	    panic("opening file %s acked! (%s)\n", nm, strerror(errno));
	cf->off_file = in.ofs;
	free(nm);
    } 
	   
    /*
     * mmap the new block 
     */
    flags = (cf->mode == CS_READER)? PROT_READ : PROT_WRITE; 
    cf->offset = ofs; 
    cf->size = *sz = in.size;

    /* 
     * align the mmap to the memory pagesize 
     */
    diff = (cf->offset - cf->off_file) % getpagesize();
    cf->offset -= diff; 
    cf->size += diff; 

    cf->addr = mmap(0, cf->size, flags, MAP_NOSYNC|MAP_SHARED, 
	cf->fd, cf->offset - cf->off_file);
    if (cf->addr == MAP_FAILED || cf->addr == NULL)
        panic("mmap got NULL (%s)\n", strerror(errno)); 

    assert(cf->addr + diff != NULL);

    return (cf->addr + diff);
}

void * 
csmap(int fd, off_t ofs, ssize_t * sz) 
{
    return cs_do_everything_else(fd, ofs, sz, S_REGION);
}

off_t
csseek(int fd)
{
    size_t retval;

    cs_do_everything_else(fd, 0, &retval, S_SEEK);
    return (off_t)retval;
}

/*
 * -- csclose
 *
 * Closes the file and sends the message to the server. 
 */
void
csclose(int fd)
{
    csmsg_t m;

    memset(&m, 0, sizeof(m));
    if (fd < 0 || fd >= CS_MAXCLIENTS || files[fd] == NULL) {
	logmsg(LOGWARN, "csclose: bad file descriptor %d\n", fd); 
	return;
    } 

    /* unmap the current block and close the file, if any */
    if (files[fd]->addr != NULL)
	munmap(files[fd]->addr, files[fd]->size);
    if (files[fd]->fd >= 0)
	close(files[fd]->fd); 

    /* send the release message to the hfd */
    m.type = S_CLOSE;
    m.id = files[fd]->id; 
    if (write(files[fd]->sd, (char *) &m, sizeof(m)) < 0) 
	panic("sending message to storage: %s\n", strerror(errno)); 

    free(files[fd]);
    files[fd] = NULL;
    return;
}


/* 
 * -- csgetofs
 * 
 * returns the current offset of the file. 
 */
off_t
csgetofs(int fd)
{
    if (!files[fd]) 
	return -1; 

    return files[fd]->offset;
} 
/* end of file */
