/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: util.c,v 1.31 2005/01/23 19:52:20 gianluca Exp $
 *
 * Debugging and various utility functions.
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h> 			/* va_start */
#include <string.h>
#include <errno.h>
#include <unistd.h>     
#include <dlfcn.h>
#include <sys/types.h>			/* inet_ntop */
#include <sys/socket.h>
#include <sys/un.h>			/* sockaddr unix */
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef linux
#include <netinet/ether.h> 		/* ether_ntoa */
#endif

#if USE_LT 	
#include <ltdl.h> 
#endif

#include "como.h"

extern struct _como map;	/* root of the data */


/** 
 * -- loglevel_name
 * 
 * Returns a string with the log message level. 
 * 
 * XXX It ignores the verbose flags. 
 */
char *
loglevel_name(int flags)
{
    static char s[1024];

    char *ui= flags & LOGUI ? "UI " : "";
    char *wa= flags & LOGWARN ? "WARN " : "";
    char *st= flags & LOGSTORAGE ? "STORAGE " : "";
    char *ca= flags & LOGCAPTURE ? "CAPTURE " : "";
    char *ex= flags & LOGEXPORT ? "EXPORT " : "";
    char *qu= flags & LOGQUERY ? "QUERY " : "";

    sprintf(s, "%s%s%s%s%s%s", ui, wa, st, ca, ex, qu);
    return s;
}


/** 
 * -- logmsg
 * 
 * Prints a message to stdout or sends it to the 
 * SUPERVISOR depending on the running loglevel.
 *
 */ 
void
logmsg(int flags, const char *fmt, ...)
{
    static int printit;	/* one copy per process */
    va_list ap;
    char *buf;
    char *fmt1;

    if (flags)
        printit = (map.logflags & flags);
    if (!printit)
        return;
    va_start(ap, fmt);
    if (flags != LOGUI)
        asprintf(&fmt1, "[%2s] %s", map.procname, fmt);
    else
        asprintf(&fmt1, "%s", fmt);
    vasprintf(&buf, fmt1, ap);
    if (map.supervisor_fd > 0) {
	write(map.supervisor_fd, buf, strlen(buf) + 1);
    } else {
	fprintf(stdout, "%s", buf);
	fflush(stdout);
    }
    va_end(ap);
    free(fmt1);
    free(buf);
}


/**
 * -- _epanic
 * 
 * Not to be called directly, but through panic().
 * Prints the message on LOGUI and aborts the program. 
 *
 */
void
_epanic(const char * file, const int line, const char *fmt, ...)
{
    char *fmt1, *buf;
    va_list ap;

    asprintf(&fmt1, "[%2s]  **** PANIC: (%s:%d) %s\n",
	map.procname, file, line, fmt);
    va_start(ap, fmt);
    vasprintf(&buf, fmt1, ap);
    va_end(ap);
    logmsg(LOGUI, "%s", buf);
    free(fmt1);
    free(buf);
    abort();
}


/**
 * -- _scalloc
 * 
 * Not to be called directly, but through safe_calloc()
 *
 * simple interface to calloc that handles errors
 * and returns only if the calloc call succeded. it
 * forces a termination, otherwise.
 *
 */
void *
_scalloc(const char * file, const int line, int n, size_t sz)
{
    void * v;

    v = calloc((unsigned int) n, sz);
    if (v == NULL) {
	fprintf(stderr, "[%2s]  **** PANIC: calloc (%s:%d)\n", 
	    map.procname, file, line);
	abort();
    }

    return v;
}


/**          
 * -- _srealloc
 *
 * Not to be called directly, but through safe_realloc()
 *
 * simple interface to realloc that handles errors
 * and returns only if the realloc call succeded. it
 * forces a termination, otherwise.
 *   
 */ 
void *
_srealloc(const char * file, const int line, void * ptr, size_t sz)
{
    void * v;
        
    v = realloc(ptr, sz);
    if (v == NULL) {
        fprintf(stderr, "[%2s]  **** PANIC: calloc (%s:%d)\n",
            map.procname, file, line);
        abort();
    }
    
    return v;
}


#ifdef USE_LT
/* 
 * -- lt_load_object
 * 
 * Like load_object just uses liblt... 
 */
void *
lt_load_object(char *base_name, char *symbol)
{
    lt_dlhandle handle;
    lt_ptr sym;

    if (1 == lt_dlinit()) {
    	panic("could not initialise libtool");
    }

    handle = lt_dlopenext(base_name);
    if (handle == NULL) {
        logmsg(LOGCONFIG, "lt_dlopen(%s) error [%s]\n",
                base_name, lt_dlerror());
        return NULL;
    }

    sym = lt_dlsym(handle, symbol);
    if (sym == NULL) {
        logmsg(LOGCONFIG, "shared object %s missing '%s' (%s)\n",
                base_name, symbol, lt_dlerror());
        lt_dlclose(handle);
        return NULL;
    }
    logmsg(LOGCONFIG, "loaded shared object %s\n", base_name);

    return (void *) sym;
}
#endif


/**
 * -- load_object
 *
 * dynamically links a library in. used for all modules as
 * well as for the filters.
 * 
 * XXX in the future we may want to use it for sniffers as well.
 *
 */
void *
load_object(char *base_name, char *symbol)
{
    void *handle, *sym;

#ifdef USE_LT 		/* use lt... XXX do we keep this? -gianluca */ 
    return lt_load_object(base_name, symbol); 
#endif

    handle = dlopen(base_name, RTLD_NOW);
    if (handle == NULL) {
        logmsg(LOGCONFIG, "dlopen(%s, RTLD_NOW) error [%s]\n",
                base_name, dlerror());
        return NULL;
    }
    sym = dlsym(handle, symbol);
    if (sym == NULL) {
        logmsg(LOGCONFIG, "module %s missing '%s' (%s)\n",
                base_name, symbol, dlerror());
        dlclose(handle);
        return NULL;
    }
    logmsg(LOGCONFIG, "loaded shared object %s\n", base_name);

    return (void *) sym;
}


/**
 * -- add_char
 *
 * add a character to a buffer, reallocating it if necessary
 *
 */
char *
add_char(char *buf, int c, uint *dst, uint *len)
{
    logmsg(V_LOGCONFIG, "add_char '%c' dst %d len %d\n", c, *dst, *len);
    if (*dst >= *len) {
        if (*len < 1024)
            *len = 2* *len + 512;
        else
            *len += 2048;
        buf = (buf == NULL) ? malloc(*len) : realloc(buf, *len);
        logmsg(V_LOGCONFIG, "realloc config buf dst %d len %d\n", *dst, *len);
        if (buf == NULL)
            panic("parse buffer realloc failed at size %d\n", *len);
    }
    buf[ (*dst)++ ] = (char)c;
    return buf;
}


/** 
 * -- print_pkt
 * 
 * Packet pretty printing...
 */
void
print_pkt(pkt_t *hdr, int have_mac)
{
    static int cnt = 0;
    char dst[20], src[20];

    if (cnt-- == 0) {
	cnt = 40;
	logmsg(V_LOGDEBUG, "%s%s%s%s",
	    "tv_sec.tv_usec length",
	    " _destination-mac_ ____source-mac___ type",
	    " vhl tos len _id_ off ttl p  sum_",
	    " ____src_ip____  ____dst_ip____\n");
    }
    //logmsg(LOGDEBUG, "%6ld.%06ld: [%4d]",
	//    hdr->ts.tv_sec % 1000000,
	//    hdr->ts.tv_usec,
    logmsg(LOGDEBUG, "%12ld: [%4d]",
       hdr->ts,
       hdr->len
    );

    if (have_mac) { /* XXX hack */
	logmsg(0, " %s",
	    ether_ntoa((struct  ether_addr *)hdr->mach.dst) );
	logmsg(0, " %s %04x",
	    ether_ntoa((struct  ether_addr *)hdr->mach.src),
	    H16(hdr->mach.type)
	);

	if (H16(hdr->mach.type) != ETHERTYPE_IP)
        goto done;
    }

    logmsg(0, "  %02x %02x %04x %04x %04x %02x %02x %04x %-16s -> %-16s",
	    hdr->ih.vhl,
	    hdr->ih.tos,
	    H16(hdr->ih.len),
	    hdr->ih.id,
	    H16(hdr->ih.ofs),
	    hdr->ih.ttl,
	    hdr->ih.proto,
	    H16(hdr->ih.cksum),
	    inet_ntop(AF_INET, &hdr->ih.src_ip, src, sizeof(src)),
	    inet_ntop(AF_INET, &hdr->ih.dst_ip, dst, sizeof(dst))
    );
    switch (hdr->ih.proto) {
    case 6:	/* tcp */ {
	logmsg(0, " %5d -> %5d", H16(hdr->p.tcph.src_port),
		H16(hdr->p.tcph.dst_port));
	}
	break;
    }
done:
    logmsg(0, "\n");
    return;
}

/* Store the PID - makes attaching the debugger less hassle */
void
createPIDfile(char * procname)
{
    char *pidfilename;
    FILE *pidfile;

    asprintf(&pidfilename, "%s/como-%s.pid", map.workdir, procname);
    pidfile = fopen(pidfilename, "w");
    if (NULL != pidfile) {
	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
    }
    free(pidfilename);
}

/*
 * Create a unix domain socket, in client or server mode.
 * XXX If the pathname does not start with / or ., we prepend map.workdir
 */
int
create_socket(const char *path, const int server)
{
    struct sockaddr_un sun;
    int i = socket(AF_UNIX, SOCK_STREAM, 0);
    int r;
    char *buf = NULL;

    if (path[0] != '/' && path[0] != '.')
	asprintf(&buf, "%s/%s", map.workdir, path);
    path = buf;
    bzero(&sun, sizeof(sun));
    sun.sun_family = AF_UNIX;
    strncpy(sun.sun_path, path, sizeof(sun.sun_path));
    if (server) {
	r = bind(i, (struct sockaddr *) &sun, sizeof(sun));
	if (r < 0)
	    panic("create_socket: cannot bind [%s] %d\n", path, r);
	listen(i, 10);
    } else { /* client mode */
	int done, retries;
	for (done = 0, retries = 0; !done && retries < 10; retries++) {
	    r = connect(i, (struct sockaddr *) &sun, sizeof(sun));
	    if (r == 0)
		done = 1;
	}
	if (!done)
	    panic("create_socket: cannot connect [%s] %d\n", path, i);
    }
    logmsg(LOGDEBUG, "create_socket %s [%s] %d\n",
	server ? "SERVER":"CLIENT", path, i);
    if (buf)
	free(buf);
    return i;
}


#if 0 

  // XXX this is never used in any part of the code.
  //     we should decide what to do with it. -gianluca

/*
 * op == 0: set counter to value
 * op != 0: add value to counter
 */
void
count_set(int off, int op, int value, char *name)
{
    ctr_t *ctr = &map.stats.base,
    *last = &map.last_stats.base,
    *t = &map.stats_times.base;

    ctr[off] = op ? ctr[off] + (uint)value : (uint)value;
    if ((uint)map.now.tv_sec != t[off]) {
        t[off] = map.now.tv_sec;
        logmsg(V_LOGDEBUG, "COUNT %8d [%8d] %s\n",
            ctr[off], ctr[off] - last[off], name);
        fflush(stdout);
        last[off] = ctr[off];
    }
}
#endif

/* end of file */
