/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: stdpkt.h,v 1.3 2005/01/23 19:52:20 gianluca Exp $
 */

#ifndef _COMO_STDPKT_H
#define _COMO_STDPKT_H

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/ethernet.h>       /* ETHERTYPE_IP */
#include <netinet/in.h>
#include <arpa/inet.h>          /* inet_ntop */


/*
 * Object types
 */
typedef struct _como_pkt        pkt_t;  /* packet descriptor */

/* typedefs for entries in network format. C99 types are used for native.
 * This makes the code a bit more boring to write, but safer in that
 * the compiler can catch error for us.
 * Access to network fields must be mediated by the 'N16()' and N32() macros
 * so it is easier to spot violations (never access explicitly
 * the field names).
 * Also never use explicitly the ntoh*(), hton*() macros.
 */
#define	N16(x)	((x).__x16)
#define	N32(x)	((x).__x32)
#define	H16(x)	(ntohs(N16(x)))
#define	H32(x)	(ntohl(N32(x)))

struct _n32_t {
    uint32_t __x32;
};

struct _n16_t {
    uint16_t __x16;
};
typedef	struct _n32_t	n32_t;	/* network format */
typedef	struct _n16_t	n16_t;	/* network format */


struct _como_machdr {
    char	dst[6];
    char	src[6];
    n16_t	type;
};

struct _como_iphdr {	/* possibly have __packed */
    uint8_t	vhl;
    uint8_t	tos;
    n16_t	len;
    n16_t	id;
    n16_t	ofs;	/* and flags */
    uint8_t	ttl;
    uint8_t	proto;
    n16_t	cksum;
    n32_t	src_ip;
    n32_t	dst_ip;
};

struct _como_tcphdr {
    n16_t	src_port;
    n16_t	dst_port;
    n32_t	seq;
    n32_t	ack;
    uint8_t	hlen;
    uint8_t	flags;
    n16_t	win;
    n16_t	cksum;
    n16_t	urg;
};

struct _como_udphdr {
    n16_t	src_port;
    n16_t	dst_port;
    n16_t	len;
    n16_t	cksum;
};

struct _como_icmphdr {
};

/*
 * macros to use for packet header fields
 * There are more macros in modules/generic_filter.c
 */

#define IP(field)	(pkt->ih.field)
#define TCP(field)	(pkt->p.tcph.field)
#define UDP(field)	(pkt->p.udph.field)


/*
 * struct _como_pkt (pkt_t) is the structure describing a packet
 * passed around the capture module.
 */

struct _como_pkt {
    uint64_t ts;                /* timestamp */
    uint32_t caplen;            /* capture length */
    uint32_t len;               /* length on the wire */
    uint32_t n_frags;           /* how many fragments ? */
    uint8_t dir;                /* traffic direction, if applicable */
    struct _como_machdr mach;   /* the mac header */
    struct _como_iphdr ih;      /* the ip header */
    union {
        struct _como_tcphdr tcph;
        struct _como_udphdr udph;
        struct _como_icmphdr icmph;
    } p;
};

/* 
 * timestamp macros 
 *
 * Convert times into NTP format timestamps -- 64 bits with
 * the high 32 for seconds and the low 32 for the fractional part.
 */
#define TS2SEC(ts) 	((uint32_t) ((ts) >> 32))
#define TS2MSEC(ts)	((u_int32_t) ((((ts) & 0xffffffff) * 1000) >> 32))
#define TS2USEC(ts)	((u_int32_t) ((((ts) & 0xffffffff) * 1000000) >> 32))
#define TIME2TS(s, u)	((((u_int64_t) (s)) << 32) + \
			((((u_int64_t) (u)) << 32) / 1000000))

#endif /* _COMO_STDPKT_H */
