/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: application.c,v 1.14 2005/01/24 12:27:26 dvercell Exp $
 */

/*
 * Application module
 *
 * This module computes the number of packets and bytes per application.
 * Output: packets/bytes per application over 1s intervals
 *
 *  [DEPRECATED] Collector formats for this module are:
 *  . "%u %hhu %llu %llu" (ts, app_code, bytes, packets)
 *
 * XXX Protocol is not considered yet to decide the application
 *
 */

#include <errno.h>
#include <time.h>

#include "como.h"
#include "module.h"

#define FLOWDESC    struct _app_stat
FLOWDESC {
    rec_t    fh;
    uint32_t ts;
    uint32_t app;
    uint64_t bytes;
    uint64_t pkts;
};

/*
 * List of known applications
 * Application code means also application priority
 */
#define WEB		0x00
#define TERMINAL	0x01	/* ssh, telnet, etc. */
#define EMAIL		0x03	/* email, news, etc. */
#define NETWORKDATA	0x07	/* dns, ntp, netbios, nfs, ftp, etc. */
#define STREAM		0x0f
#define GAMES		0x1f
#define SHARING		0x3f
#define UNKNOWN		0xff    

#define APP(map, sport, dport)	(((map)[sport]) & ((map)[dport]))

static uint8_t * port2app;	/* XXX this could be a static array but
				 *     we go thru init just to exercise
				 *     some code in /base
			 	 */

#define MAXAPPS         256
static char * app2str[MAXAPPS];

static int
init(void *mem, size_t msize)
{
    uint i;

    if (msize < 65536) {
	logmsg(LOGWARN, "need 64K of private memory, have just %d\n", msize);
	return ENOMEM; 
    } 

    memset(mem, UNKNOWN, msize);
    port2app = (uint8_t *) mem;

    /* initialize the port-to-application mapping array */
    port2app[80]    = WEB;
    port2app[443]   = WEB;
    port2app[8080]  = WEB;
    port2app[3128]  = WEB;             // SQUID
    port2app[3130]  = WEB;             // SQUID
    port2app[9090]  = WEB;             // AUTOPROXY 

    port2app[6688]  = SHARING;
    port2app[6697]  = SHARING;
    port2app[6699]  = SHARING;
    port2app[4329]  = SHARING;
    port2app[4444]  = SHARING;
    port2app[5555]  = SHARING;
    port2app[6666]  = SHARING;
    port2app[7777]  = SHARING;
    port2app[6346]  = SHARING;
    port2app[1214]  = SHARING;
    for (i = 4000; i <= 4999; i++) {
        port2app[i] = SHARING;
    }

    port2app[22]    = TERMINAL;         // SSH
    port2app[23]    = TERMINAL;         // TELNET
    port2app[992]   = TERMINAL;         // TELNET

    port2app[25]    = EMAIL;             //SMTP
    port2app[465]   = EMAIL;             //SMTPS
    port2app[109]   = EMAIL;             //POP
    port2app[110]   = EMAIL;             //POP
    port2app[995]   = EMAIL;             //POP
    port2app[143]   = EMAIL;             //IMAP
    port2app[220]   = EMAIL;             //IMAP
    port2app[993]   = EMAIL;             //IMAP
    port2app[119]   = EMAIL;             //NNTP
    port2app[563]   = EMAIL;             //NNTP

    port2app[20]    = NETWORKDATA;	// FTP
    port2app[21]    = NETWORKDATA;	// FTP
    port2app[989]   = NETWORKDATA;	// FTP
    port2app[990]   = NETWORKDATA;	// FTP
    port2app[53]    = NETWORKDATA;      //DNS
    port2app[161]   = NETWORKDATA;          //SNMP
    port2app[162]   = NETWORKDATA;          //SNMP
    port2app[123]   = NETWORKDATA;          //NTP
    port2app[873]   = NETWORKDATA;    	//RSYNC
    port2app[1110]  = NETWORKDATA;     //NFS
    port2app[2049]  = NETWORKDATA;     //NFS
    port2app[135]   = NETWORKDATA;        //NETBIOS
    port2app[137]   = NETWORKDATA;        //NETBIOS
    port2app[138]   = NETWORKDATA;        //NETBIOS
    port2app[139]   = NETWORKDATA;        //NETBIOS
    port2app[445]   = NETWORKDATA;        //NETBIOS
    port2app[568]   = NETWORKDATA;        //NETBIOS
    port2app[569]   = NETWORKDATA;        //NETBIOS
    port2app[1512]  = NETWORKDATA;        //NETBIOS
    port2app[311]   = NETWORKDATA;        //APPLETALK
    port2app[387]   = NETWORKDATA;        //APPLETALK
    port2app[548]   = NETWORKDATA;        //APPLETALK

    port2app[4662]  = SHARING;              //E_DONKEY
    port2app[7070]  = SHARING;              //E_DONKEY
    port2app[1214]  = SHARING;              //FASTTRACK
    port2app[6346]  = SHARING;              //GNUTELLA
    port2app[412]   = SHARING;              //DIRECT_CONNECT
    port2app[5000]  = SHARING;              //IMESH_CONTROL
    port2app[4329]  = SHARING;              //IMESH_DATA
    port2app[6574]  = SHARING;              //ROMNET
    port2app[8311]  = SHARING;              //SCOUR_EX
    port2app[5500]  = SHARING;              //HOTLINE
    port2app[5501]  = SHARING;              //HOTLINE
    
    memset(app2str, 0, sizeof(char*) * MAXAPPS);
    app2str[WEB]         = "WEB";
    app2str[TERMINAL]    = "TRM";
    app2str[EMAIL]       = "EML";
    app2str[NETWORKDATA] = "NET";
    app2str[STREAM]      = "STR";
    app2str[GAMES]       = "GAM";
    app2str[SHARING]     = "P2P";
    app2str[UNKNOWN]     = "??";

    return 0;
}

static uint32_t
hash(pkt_t *pkt)
{
    return APP(port2app, H16(TCP(src_port)), H16(TCP(dst_port))); 
}

static int
match(pkt_t *pkt, rec_t *fh)
{
    FLOWDESC *x = F(fh);
    uint8_t code = APP(port2app, H16(TCP(src_port)), H16(TCP(dst_port)));
    return (x->app == (uint32_t) code); 
}

static int
update(pkt_t *pkt, rec_t *fh, int isnew)
{
    FLOWDESC *x = F(fh);

    if (isnew) {
        x->bytes = 0;
        x->pkts = 0;
    }

    x->app = (uint32_t) APP(port2app, H16(TCP(src_port)), H16(TCP(dst_port)));
    x->ts = TS2SEC(pkt->ts);
    x->bytes += H16(IP(len));
    x->pkts++;

    return 0;
}

#define REC_SIZE        24      /* actual size of the record on disk */

static ssize_t
store(rec_t *fh, char *buf, size_t len)
{
    FLOWDESC *x = F(fh);
    
    if (len < REC_SIZE)
        return -1;

    PUTH32(buf,      x->ts);
    PUTH32(buf + 4,  x->app);
    PUTH64(buf + 8,  x->bytes);
    PUTH64(buf + 16, x->pkts);

    return REC_SIZE;
}

static size_t
load(char * buf, size_t len, timestamp_t * ts)
{
    if (len < REC_SIZE) {
        ts = 0;
        return 0; 
    }

    *ts = TIME2TS(ntohl(*(uint32_t *)buf), 0);
    return REC_SIZE;
}

static char *
print(char *buf)
{
    FLOWDESC *x = F(buf - sizeof(rec_t));
    static char s[2048];
    
    time_t ts = (time_t) ntohl(x->ts);
    
    sprintf(s, "Ts: %.24s App: %3s Bytes (1s): %8llu Packets (1s): %llu\n",
	asctime(localtime(&ts)), app2str[ntohl(x->app)], 
	NTOHLL(x->bytes), NTOHLL(x->pkts));
    return s;
};


callbacks_t callbacks = {
    sizeof(FLOWDESC),
    0,
    init,
    NULL,
    hash,
    match,
    update,
    NULL,
    NULL,
    NULL,
    NULL,
    store,
    load,
    print
};
