/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: topdest.c,v 1.12 2005/01/24 16:45:08 dvercell Exp $
 */

/*
 * Top-N destinations Module.
 *
 * This module computes the top-N IP destination addresses
 *
 * [DEPRECATED] Collector formats for this module are:
 *  . "%u %a %llu %llu" (ts, dst_ip, bytes, pkts)
 *
 *
 */

#include <time.h>
#include "como.h"
#include "module.h"

#define TOPN        10			/* Top-10 destinations */
#define EXPORT_IVL  TIME2TS(5,0)	/* export interval (seconds) */

#define FLOWDESC	struct _topdest_stat
#define EFLOWDESC	FLOWDESC

FLOWDESC {
    rec_t fh;
    uint32_t ts;	/* timestamp of last packet */
    n32_t dst_ip;	/* destination IP address */
    uint64_t bytes;	/* number of bytes */
    uint64_t pkts;	/* number of packets */
};

static timestamp_t last_export = 0; 


static uint32_t
hash(pkt_t *pkt)
{
    return (N32(IP(dst_ip))); 
}

static int
match(pkt_t *pkt, rec_t *fh)
{
    FLOWDESC *x = F(fh);
    return (N32(IP(dst_ip)) == N32(x->dst_ip));
}

static int
update(pkt_t *pkt, rec_t *fh, int isnew)
{
    FLOWDESC *x = F(fh);

    if (isnew) {
        x->dst_ip = IP(dst_ip);
        x->bytes = 0;
        x->pkts = 0;
    }

    x->ts = TS2SEC(pkt->ts);
    x->bytes += H16(IP(len));
    x->pkts ++;

    return 0;
}

static int
ematch(rec_t *efh, rec_t *fh)
{
    FLOWDESC *x = F(fh);
    EFLOWDESC *ex = EF(efh);

    return (N32(x->dst_ip) == N32(ex->dst_ip));
}

static int
export(rec_t *efh, rec_t *fh, int isnew)
{
    FLOWDESC *x = F(fh);
    EFLOWDESC *ex = EF(efh);

    if (isnew) {
        ex->dst_ip = x->dst_ip;
        ex->bytes = 0;
        ex->pkts = 0;
    }

    ex->ts = TS2SEC(last_export);
    ex->bytes += x->bytes;
    ex->pkts += x->pkts;

    return 0;
}

static int
compare(const void *efh1, const void *efh2)
{
    EFLOWDESC *ex1 = CMPEF(efh1);
    EFLOWDESC *ex2 = CMPEF(efh2);

    return ((ex1->bytes > ex2->bytes)? -1 : 1);
}

static int
action(rec_t *efh, timestamp_t current_time, int count)
{
    if (efh == NULL) { 
	/* this is the action for the entire table. 
	 * check if it is time to export the table. 
	 * if not stop. 
	 */
	if (current_time - last_export < EXPORT_IVL) 
	    return ACT_STOP;		/* too early */

	last_export = current_time; 
	return ACT_GO; 		/* dump the records */
    }

    return (count < TOPN)? ACT_STORE|ACT_DISCARD : ACT_DISCARD; 
}

#define REC_SIZE    24			/* size of export record */

static ssize_t
store(rec_t *efh, char *buf, size_t len)
{
    EFLOWDESC *ex = EF(efh);
    
    if (len < REC_SIZE)
        return -1;

    PUTH32(buf, ex->ts);
    PUTN32(buf + 4, N32(ex->dst_ip));
    PUTH64(buf + 8, ex->bytes);
    PUTH64(buf + 16, ex->pkts);

    return REC_SIZE;
}

static size_t
load(char *buf, size_t len, timestamp_t *ts)
{
    if (len < REC_SIZE) {
        ts = 0;
        return 0; 
    }

    *ts = TIME2TS(ntohl(*(uint32_t *)buf), 0);
    return REC_SIZE;
}

static char *
print(char *buf)
{
    FLOWDESC *x = F(buf - sizeof(rec_t));
    static char s[2048];
    struct in_addr addr;
    time_t ts = (time_t) ntohl(x->ts);

    addr.s_addr = N32(x->dst_ip);
    
    sprintf(s, "Ts: %.24s Dest ip: %15s Bytes (1s): %llu\tPkts (1s): %llu\n",
	asctime(localtime(&ts)), inet_ntoa(addr), 
	NTOHLL(x->bytes), NTOHLL(x->pkts));
    return s;
};


callbacks_t callbacks = {
    sizeof(FLOWDESC),
    sizeof(EFLOWDESC),
    NULL,
    NULL,
    hash,
    match,
    update,
    ematch,
    export,
    compare,
    action,
    store,
    load,
    print
};
