/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: tuple.c,v 1.14 2005/01/25 22:07:20 gianluca Exp $
 */

/*
 * 5-tuple flow classifier
 *
 * This module computes 5-tuple flows. Every capture period it 
 * stores the 5-tuple that have been active (together with the 
 * bytes they have sent).  
 *
 */

#include <time.h>
#include "como.h"
#include "module.h"

#define FLOWDESC    struct _tuple_stat

FLOWDESC {
    rec_t fh;
    uint32_t ts; 
    n32_t src_ip;
    n32_t dst_ip;
    n16_t src_port;
    n16_t dst_port;
    uint8_t proto;
    char padding[3];
    uint64_t bytes;
    uint64_t pkts;
};


static uint32_t
hash(pkt_t *pkt)
{
    return (N32(IP(src_ip)) ^ N32(IP(dst_ip)) ^
             (N16(TCP(src_port)) << 3) ^ (N16(TCP(dst_port)) << 3));
}

static int
match(pkt_t *pkt, rec_t *fh)
{
    FLOWDESC *x = F(fh);

    return (
         N32(IP(src_ip)) == N32(x->src_ip) &&
         N32(IP(dst_ip)) == N32(x->dst_ip) &&
         N16(TCP(src_port)) == N16(x->src_port) &&
         N16(TCP(dst_port)) == N16(x->dst_port) &&
         IP(proto) == x->proto
    );
}

static int
update(pkt_t *pkt, rec_t *fh, int isnew)
{
    FLOWDESC *x = F(fh);

    if (isnew) {
        x->src_ip = IP(src_ip);
        x->dst_ip = IP(dst_ip);
        x->src_port = TCP(src_port);
        x->dst_port = TCP(dst_port);
        x->proto = IP(proto);
	x->bytes = 0;
	x->pkts = 0;
	x->ts = TS2SEC(pkt->ts); 
    }

    x->bytes += H16(IP(len));
    x->pkts++;

    return 0;
}

#define REC_SIZE    	(sizeof(FLOWDESC) - sizeof(rec_t))

static ssize_t
store(rec_t *efh, char *buf, size_t len)
{
    FLOWDESC *x = F(efh);
    
    if (len < REC_SIZE)
        return -1;

    PUTH32(buf, x->ts);
    PUTN32(buf + 4, N32(x->src_ip));
    PUTN32(buf + 8, N32(x->dst_ip));
    PUTN16(buf + 12, N16(x->src_port));
    PUTN16(buf + 14, N16(x->dst_port));
    PUTH8(buf + 16, x->proto);
    PUTH8(buf + 17, x->padding[0]);
    PUTH8(buf + 18, x->padding[1]);
    PUTH8(buf + 19, x->padding[2]);
    PUTH64(buf + 20, x->bytes);
    PUTH64(buf + 28, x->pkts);

    return REC_SIZE;
}

static size_t
load(char * buf, size_t len, timestamp_t * ts)
{
    if (len < REC_SIZE) {
        ts = 0;
        return 0;
    }

    *ts = NTOHLL(*(uint64_t *)buf);
    return REC_SIZE;
}

static char *
print(char *buf)
{
    FLOWDESC *x = F(buf - sizeof(rec_t));
    static char s[2048];
    char src[20], dst[20];
    struct in_addr saddr, daddr;
    time_t ts;

    ts = (time_t)ntohl(x->ts);
    saddr.s_addr = N32(x->src_ip);
    daddr.s_addr = N32(x->dst_ip);
    sprintf(src, "%s", inet_ntoa(saddr));
    sprintf(dst, "%s", inet_ntoa(daddr)); 

    sprintf(s, "Time: %.24s %15s %15s sport %5u dport %5u proto: %3u "
                "Bytes: %8llu Packets: %8llu",
                asctime(localtime(&ts)), src, dst, 
                (uint) H16(x->src_port), (uint) H16(x->dst_port), 
		(uint) x->proto, NTOHLL(x->bytes), NTOHLL(x->pkts));
    return s;
};


callbacks_t callbacks = {
    sizeof(FLOWDESC),
    0, 
    NULL,
    NULL,
    hash,
    match,
    update,
    NULL,
    NULL,
    NULL,
    NULL,
    store,
    load,
    print
};
