/*
 * Copyright (c) 2004 Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: utilization.c,v 1.15 2005/01/25 17:59:52 gianluca Exp $
 */


/*
 * Utilization Module.
 *
 * This module computes the link utilization using the following four metrics:
 *
 *  . bits/sec averaged over 1s intervals;
 *  . high watermark (i.e., max utilization in 100ms interval) over 1s periods;
 *
 * [DEPRECATED]
 * Collector formats for this module are:
 *
 *  . "%u %llu %llu" (ts, bits/sec, watermark)
 *
 */

#include <time.h>
#include "como.h"
#include "module.h"

#define WATERMARK_IVL   TIME2TS(0, 100000)     /* watermark ivl */

#define FLOWDESC        struct _util_stat

FLOWDESC {
    rec_t fh;
    uint32_t ts;
    uint32_t bytes;
    uint32_t hi_watermark;
    uint32_t bytes_ivl;
    timestamp_t last_watermark;
};


static int
update(pkt_t *pkt, rec_t *fh, int isnew)
{
    FLOWDESC *x = F(fh);

    if (isnew) {
        x->bytes = 0;
        x->bytes_ivl = 0;
        x->hi_watermark = 0;
        x->last_watermark = pkt->ts;
    }

    x->ts = TS2SEC(pkt->ts);
    x->bytes += H16(IP(len));

    if (pkt->ts - x->last_watermark > WATERMARK_IVL) {
        x->last_watermark = pkt->ts;
        x->bytes_ivl = 0;
    }

    x->bytes_ivl += H16(IP(len));
    if (x->bytes_ivl > x->hi_watermark)
        x->hi_watermark = x->bytes_ivl;

    return 0;
}


#define REC_SIZE        12      /* size of capture/export record */

static ssize_t
store(rec_t *efh, char * buf, size_t len)
{
    FLOWDESC *x = F(efh);

    if (len < REC_SIZE)
        return -1;
    
    PUTH32(buf, x->ts);
    PUTH32(buf + 4, x->bytes);
    PUTH32(buf + 8, x->hi_watermark);
    return REC_SIZE;
}

static size_t
load(char *buf, size_t len, timestamp_t *ts)
{
    if (len < REC_SIZE) {
        ts = 0;
        return 0;
    }

    *ts = TIME2TS(ntohl(*(uint32_t *)buf), 0);
    return REC_SIZE;
}

static char * 
print(char *buf)
{
   static char s[2048];
   FLOWDESC *x = F(buf - sizeof(rec_t));
   time_t ts = (time_t) ntohl(x->ts);

   sprintf(s, "Time: %.24s Bytes: %10u HighWaterMark: %10u", 
	asctime(localtime(&ts)), (uint) ntohl(x->bytes), 
	(uint) ntohl(x->hi_watermark));

   return s;
}

callbacks_t callbacks = {
    sizeof(FLOWDESC),
    0,
    NULL,
    NULL,
    NULL,
    NULL,
    update,
    NULL,
    NULL,
    NULL,
    NULL,
    store,
    load,
    print
};
